# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
import importlib
from itertools import chain
import json
import os
import sys
import time
import urllib.request

import click
from tqdm import tqdm

from fastwarc.stream_io import FileStream, StreamError, FastWARCError, PythonIOStreamAdapter
from fastwarc.warc import ArchiveIterator, WarcRecordType
from fastwarc.tools import CompressionAlg, detect_compression_algorithm, wrap_warc_stream, \
    recompress_warc_interactive, verify_digests


def exception_handler(exctype, value, _):
    if exctype == FastWARCError:
        click.echo(str(value), err=True)
        sys.exit(1)

    raise value


sys.excepthook = exception_handler


@click.group(context_settings=dict(help_option_names=['-h', '--help']))
def main():
    """FastWARC Command Line Interface."""
    return 0


def _human_readable_bytes(byte_num):
    for unit in ['bytes', 'KiB', 'MiB', 'GiB', 'TiB']:
        if byte_num <= 1024 or unit == 'TiB':
            if int(byte_num) == byte_num:
                return f'{byte_num:d} {unit}'
            return f'{byte_num:.2f} {unit}'
        byte_num /= 1024


@main.command()
@click.argument('infile', type=click.Path(dir_okay=False, exists=True))
@click.argument('outfile', type=click.Path(dir_okay=False, exists=False))
@click.option('-c', '--compress-alg', type=click.Choice(['gzip', 'lz4', 'uncompressed', 'auto']),
              default='auto', show_default=True, help='Compression algorithm to use for output file')
@click.option('-d', '--decompress-alg', type=click.Choice(['gzip', 'lz4', 'uncompressed', 'auto']),
              default='auto', show_default=True,
              help='Decompression algorithm for decoding input file (auto tries to detect based on file extension)')
@click.option('-l', '--compress-level', type=int, default=None, help='Compression level (defaults to max)')
@click.option('-q', '--quiet', is_flag=True, help='Do not print progress information')
def recompress(infile, outfile, compress_alg, decompress_alg, compress_level, quiet):
    """
    Recompress a WARC file.

    This command allows you to recompress a WARC file if it is uncompressed or not compressed
    properly at the record-level if or you want to recompress a GZip WARC as LZ4 or vice versa.
    """

    compress_alg = getattr(CompressionAlg, compress_alg)
    decompress_alg = getattr(CompressionAlg, decompress_alg)

    comp_args = {}
    if compress_level is None:
        if compress_alg == CompressionAlg.gzip:
            comp_args['compression_level'] = 9
        elif compress_alg == CompressionAlg.lz4:
            comp_args['compression_level'] = 12

    bytes_written = 0
    num = 0
    start = time.time()
    try:
        for _, b in tqdm(recompress_warc_interactive(infile, outfile, decompress_alg, compress_alg, **comp_args),
                         desc='Recompressing WARC file', unit=' record(s)', leave=False,
                         disable=quiet, mininterval=0.2):
            num += 1
            bytes_written += b

        if not quiet:
            click.echo('Recompression completed.')
    except KeyboardInterrupt:
        if not quiet:
            click.echo('Recompression aborted.')
            sys.exit(1)
    finally:
        if not quiet and num > 0:
            click.echo(f'  - Records recompressed: {num}')
            click.echo(f'  - Bytes written: {_human_readable_bytes(bytes_written)}')
            click.echo(f'  - Completed in: {time.time() - start:.02f} seconds')


@main.command()
@click.argument('infile', type=click.Path(dir_okay=False, exists=True))
@click.option('-d', '--decompress-alg', type=click.Choice(['gzip', 'lz4', 'uncompressed', 'auto']),
              default='auto', show_default=True, help='Decompression algorithm')
@click.option('-p', '--verify-payloads', is_flag=True, help='Also verify payload digests')
@click.option('-q', '--quiet', is_flag=True, help='Do not print progress information')
@click.option('-o', '--output', type=click.File('w'), help='Output file with verification details')
def check(infile, decompress_alg, verify_payloads, quiet, output):
    """
    Verify WARC record consistency.

    Check digests of all records in a WARC file and print a summary. You can verify all block and
    payload digests in the given WARC file and print a summary of all corrupted and (optionally)
    all intact records.

    The command will exit with a non-zero exit code if at least one record fails verification.
    """

    decompress_alg = getattr(CompressionAlg, decompress_alg)

    failed_digests = []
    num_total = 0
    num_no_digest = 0
    try:
        desc_tpl = 'Verifying digests ({} failed)'
        pbar = tqdm(verify_digests(infile, verify_payloads, decompress_alg),
                    desc=desc_tpl.format(0), unit=' record(s)', leave=False, disable=quiet, mininterval=0.2)

        for v in pbar:
            num_total += 1

            status = 'OK'
            if v['block_digest_ok'] is False:
                status = 'FAIL'
            elif v['block_digest_ok'] is None:
                status = 'NO_DIGEST'

            if 'payload_digest_ok' in v:
                if v['payload_digest_ok'] is True:
                    status += ', PAYLOAD_OK'
                elif v['payload_digest_ok'] is False:
                    status += ', PAYLOAD_FAIL'
                elif v['payload_digest_ok'] is None:
                    status += ', PAYLOAD_NO_DIGEST'

            if 'FAIL' in status:
                failed_digests.append(v['record_id'])
                pbar.set_description(desc_tpl.format(len(failed_digests)))
            elif 'NO_DIGEST' in status and 'OK' not in status:
                num_no_digest += 1

            if output:
                output.write(f'{v["record_id"]}: {status}\n')

        if output:
            output.close()

    except KeyboardInterrupt:
        if not quiet:
            click.echo('Verification aborted.')
            sys.exit(1)
    finally:
        if not quiet and num_total > 0:
            if not failed_digests:
                click.echo(f'{num_total - num_no_digest} records were verified successfully.')
                if num_no_digest:
                    click.echo(f'{num_no_digest} records were skipped without digest.')
            else:
                click.echo('Failed records:')
                click.echo('===============')
                for rec in failed_digests:
                    click.echo(rec)
                sys.exit(1)


@main.command()
@click.argument('infile', type=click.Path(dir_okay=False, exists=True))
@click.argument('offset', type=int)
@click.option('-o', '--output', type=click.File('wb'), default=sys.stdout.buffer,
              help='Output file, default is stdout')
@click.option('--payload', is_flag=True,
              help='Output only record payload (transfer and/or content encoding are preserved')
@click.option('--headers', is_flag=True, help='Output only record (and HTTP) headers')
def extract(infile, offset, output, payload, headers):
    """
    Extract WARC record by offset.

    You can extract individual records at a given byte offset with either just
    headers, payload, or both.
    """

    compl_record = not (payload or headers)
    with open(infile, 'rb') as stream:
        stream.seek(offset)
        try:
            record = next(ArchiveIterator(stream))
            if compl_record:
                record.write(output)
            elif headers:
                record.headers.write(PythonIOStreamAdapter(output))
                output.write(b'\r\n')
                if record.is_http:
                    record.http_headers.write(PythonIOStreamAdapter(output))
                    output.write(b'\r\n')
            elif payload:
                buf = record.reader.read(4096)
                while buf:
                    output.write(buf)
                    buf = record.reader.read(4096)
        except StopIteration:
            return
        except (FastWARCError, OSError, ValueError) as e:
            click.echo(f'Failed to extract WARC record at offset {offset}: {str(e)}', err=True)


def _index_record(output, fields, preserve_multi_header, record, next_record_offset, file_name):
    idx = dict()
    for f in fields:
        f = f.strip().lower()

        if f == 'offset':
            idx[f] = str(record.stream_pos)
        elif f == 'length':
            idx[f] = str(next_record_offset - record.stream_pos)
        elif f == 'filename':
            idx[f] = file_name
        elif f == 'http:status' and record.is_http:
            idx[f] = str(record.http_headers.status_code)
        elif f.startswith('http:') and record.is_http:
            if preserve_multi_header:
                l = []
                for k, v in record.http_headers:
                    if k == f[5:]:
                        l.append(v)
                if len(l) == 1:
                    idx[f] = l[0]
                elif len(l) > 1:
                    idx[f] = l
            else:
                idx[f] = record.http_headers.get(f[5:])
        elif f in record.headers:
            idx[f] = record.headers.get(f)

    output.write(json.dumps(idx) + '\n')


@main.command()
@click.argument('infiles', type=click.Path(dir_okay=False, exists=True), nargs=-1)
@click.option('-o', '--output', type=click.File('w'), default=sys.stdout,
              help='Output file, default is stdout')
@click.option('-f', '--fields', type=str, metavar='FIELDS',
              default='offset,warc-type,warc-target-uri', show_default=True,
              help='Comma-separated list of indexed fields, eg. "offset", "length", "filename", '
                   '"http:status", "http:<http-header>", or "<warc-record-header>"')
@click.option('--preserve-multi-header', is_flag=True,
              help='Preserve multiple values of HTTP headers as JSON list')
def index(infiles, output, fields, preserve_multi_header):
    """
    Index WARC records as CDXJ.

    WARC files can be indexed to the `CDXJ <https://github.com/webrecorder/cdxj-indexer>`_
    format with a configurable set of fields.
    """
    fields = fields.split(',')
    for infile in infiles:
        with open(infile, 'rb') as stream:
            prev_record = None
            for record in ArchiveIterator(stream):
                if prev_record is not None:
                    _index_record(output, fields, preserve_multi_header, prev_record, record.stream_pos, infile)
                prev_record = record

            if prev_record is not None:
                _index_record(output, fields, preserve_multi_header, prev_record, stream.tell(), infile)


boto3 = None
botocore = None
s3 = None


def _init_s3(endpoint_url, aws_access_key,  aws_secret_key=None):
    global boto3, botocore, s3

    if s3 is not None:
        return

    try:
        boto3 = importlib.import_module('boto3')
        botocore = importlib.import_module('botocore')
    except ModuleNotFoundError:
        raise click.UsageError('Boto3 needs to be installed for S3 benchmarking.')

    if not aws_access_key:
        raise click.UsageError("s3:// URLs require '--aws-access-key' to be set.")
    if not aws_secret_key:
        aws_secret_key = getpass.getpass()

    s3 = boto3.resource('s3', endpoint_url=endpoint_url, aws_access_key_id=aws_access_key,
                        aws_secret_access_key=aws_secret_key)


# noinspection PyProtectedMember, HttpUrlsUsage
def _get_raw_stream_from_url(input_url, use_python_stream=False):
    if input_url.startswith('s3://'):
        assert s3 is not None
        try:
            s3_bucket, s3_object = input_url[5:].split('/', 1)
            stream = s3.Object(s3_bucket, s3_object).get()['Body']._raw_stream
        except botocore.exceptions.ClientError as e:
            click.echo(f'Error connecting to S3. Message: {e}.', err=True)
            sys.exit(1)

    elif input_url.startswith('http://') or input_url.startswith('https://'):
        stream = urllib.request.urlopen(input_url)

    else:
        if input_url.startswith('file://'):
            input_url = input_url[7:]
        if not os.path.isfile(input_url):
            click.echo(f"No such file or directory: '{input_url}", err=True)
            sys.exit(1)

        stream = FileStream(input_url, 'rb') if not use_python_stream else open(input_url, 'rb')

    return stream


def _expand_s3_prefixes(input_urls, endpoint_url, aws_access_key, aws_secret_key):
    urls_expanded = []
    for url in input_urls:
        if url.startswith('s3://'):
            _init_s3(endpoint_url, aws_access_key, aws_secret_key)
            s3_bucket, s3_object_prefix = url[5:].split('/', 1)
            urls_expanded.extend('/'.join(('s3:/', s3_bucket, o.key))
                                 for o in s3.Bucket(s3_bucket).objects.filter(Prefix=s3_object_prefix))
        else:
            urls_expanded.append(url)
    return urls_expanded


# noinspection PyPackageRequirements
@main.command()
@click.argument('input_url', nargs=-1)
@click.option('-d', '--decompress-alg', type=click.Choice(['gzip', 'lz4', 'uncompressed', 'auto']),
              default='auto', show_default=True, help='Decompression algorithm')
@click.option('-e', '--endpoint-url', help='S3 endpoint URL', default='https://s3.amazonaws.com', show_default=True)
@click.option('-a', '--aws-access-key', help='AWS access key for s3:// URLs')
@click.option('-s', '--aws-secret-key', help='AWS secret key for s3:// URLs (leave empty to read from STDIN)')
@click.option('--is-prefix', is_flag=True, help='Treat input URL as prefix (only for S3)')
@click.option('-p', '--use-python-stream', is_flag=True,
              help='Use slower Python I/O instead of native FileStream for local files')
@click.option('-H', '--parse-http', is_flag=True, help='Parse HTTP headers')
@click.option('-v', '--verify-digests', is_flag=True, help='Verify record block digests')
@click.option('-f', '--filter-type', type=click.Choice(['warcinfo', 'response', 'resource', 'request', 'metadata',
                                                        'revisit', 'conversation', 'continuation', 'any_type']),
              default=['any_type'], multiple=True, show_default=True, help='Filter for specific WARC record types')
@click.option('-w', '--bench-warcio', is_flag=True, help='Compare FastWARC performance with WARCIO')
def benchmark(input_url, decompress_alg, endpoint_url, aws_access_key, aws_secret_key, is_prefix, use_python_stream,
              verify_digests, parse_http, filter_type, bench_warcio):
    """
    Benchmark FastWARC read performance.

    The FastWARC CLI comes with a benchmarking tool for measuring WARC record decompression and parsing
    performance on your own machine. The benchmarking results can be compared directly with WARCIO.

    Supported WARC sources are local files, S3 and HTTP(s) URLs. Supported compression algorithms
    are GZip, LZ4, or uncompressed.

    The read benchmarking tool has additional options, such as reading WARCs directly from a remote S3 data source
    using `Boto3 <https://boto3.amazonaws.com/v1/documentation/api/latest/index.html>`_.

    See :ref:`FastWARC Benchmarks <fastwarc-benchmarks>` for more information and example benchmarking results.
    """

    if not input_url:
        raise click.UsageError('Need at least one input path.')

    if is_prefix:
        input_url = _expand_s3_prefixes(input_url, endpoint_url, aws_access_key, aws_secret_key)
    else:
        for u in input_url:
            if u.startswith('s3://'):
                _init_s3(endpoint_url, aws_access_key, aws_secret_key)
                break

    if decompress_alg == 'auto':
        decompress_alg = detect_compression_algorithm(input_url[0])
    else:
        decompress_alg = getattr(CompressionAlg, decompress_alg)

    if bench_warcio:
        if decompress_alg == CompressionAlg.lz4:
            click.echo('WARCIO does not support LZ4.', err=True)
            sys.exit(1)

        try:
            import warcio
        except ModuleNotFoundError:
            click.echo("--bench-warcio requires 'warcio' to be installed.", err=True)
            sys.exit(1)

    click.echo(f'Benchmarking read performance from {len(input_url)} input path(s)...')

    rec_type_filter = WarcRecordType.any_type
    if 'any_type' not in filter_type:
        rec_type_filter = WarcRecordType.no_type
        for t in filter_type:
            rec_type_filter |= getattr(WarcRecordType, t)

    def _bench(urls, stream_loader, lib):
        interrupted = False
        start = time.monotonic()
        num = 0

        try:
            def _load_lazy(u, l):
                yield from l(u)

            for _ in tqdm(chain(*(_load_lazy(u, stream_loader) for u in urls)),
                          desc=f'Benchmarking {lib}', unit=' records', leave=False, mininterval=0.5):
                num += 1
        except StreamError as e:
            click.echo(f'Error reading input: {e}', err=True)
            sys.exit(1)
        except KeyboardInterrupt:
            click.echo(f'Benchmark interrupted.', err=True)
            interrupted = True

        return num, time.monotonic() - start, interrupted

    def _fastwarc_iterator(f):
        s = _get_raw_stream_from_url(f, use_python_stream)
        s = wrap_warc_stream(s, 'rb', decompress_alg)
        return ArchiveIterator(s, rec_type_filter, parse_http=parse_http, verify_digests=verify_digests)

    n, t_fastwarc, interrupted = _bench(input_url, _fastwarc_iterator, 'FastWARC')
    click.echo(f'FastWARC: {n:,} records read in {t_fastwarc:.02f} seconds ({n / t_fastwarc:,.02f} records/s).')

    if interrupted:
        sys.exit(1)

    if bench_warcio:
        def _warcio_iterator(f):
            s = _get_raw_stream_from_url(f, True)
            if rec_type_filter == WarcRecordType.any_type:
                yield from warcio.ArchiveIterator(s, no_record_parse=not parse_http, check_digests=verify_digests)
            else:
                # WARCIO does not support record filtering out of the box
                for rec in warcio.ArchiveIterator(s, not parse_http):
                    if rec_type_filter & getattr(WarcRecordType, rec.rec_type) != 0:
                        yield rec

        n, t_warcio, _ = _bench(input_url, _warcio_iterator, 'WARCIO')
        click.echo(f'WARCIO:   {n:,} records read in {t_warcio:.02f} seconds ({n / t_warcio:,.02f} records/s).')
        click.echo(f'Time difference: {t_fastwarc - t_warcio:.02f} seconds, speedup: {t_warcio / t_fastwarc:,.02f}')


if __name__ == '__main__':
    main()



from libcpp.string cimport string

cdef extern from "<boost/regex.hpp>" namespace "boost::regex_constants" nogil:
    ctypedef enum flag_type:
        perl_syntax_group,
        basic_syntax_group,
        literal,
        main_option_type,
        no_bk_refs,
        no_perl_ex,
        no_mod_m,
        mod_x,
        mod_s,
        no_mod_s,
        no_char_classes,
        no_intervals,
        bk_plus_qm,
        bk_vbar,
        emacs_ex,
        no_escape_in_lists,
        newline_alt,
        no_except,
        failbit,
        icase,
        nocollate,
        collate,
        nosubs,
        save_subexpression_location,
        no_empty_expressions,
        optimize,

        basic,
        extended,
        normal,
        emacs,
        awk,
        grep,
        egrep,
        sed,
        perl,
        ECMAScript,
        JavaScript,
        JScript

    ctypedef enum match_flag_type:
        match_default,
        match_not_bol,
        match_not_eol,
        match_not_bob,
        match_not_eob,
        match_not_bow,
        match_not_eow,
        match_not_dot_newline,
        match_not_dot_null,
        match_prev_avail,
        match_init,
        match_any,
        match_not_null,
        match_continuous,
        match_partial,
        match_stop,
        match_not_initial_null,
        match_all,
        match_perl,
        match_posix,
        match_nosubs,
        match_extra,
        match_single_line,
        match_unused1,
        match_unused2,
        match_unused3,
        match_max,
        format_perl,
        format_default,
        format_sed,
        format_all,
        format_no_copy,
        format_first_only,
        format_is_if,
        format_literal,
        match_not_any

from resiliparse_inc.string_view cimport string_view
cdef extern from "<boost/regex.hpp>" namespace "boost" nogil:
    cdef cppclass basic_regex[charT]:
        basic_regex()
        basic_regex(const basic_regex& rgx)
        basic_regex(const charT* p)
        basic_regex(const charT* p, flag_type f)
        basic_regex(const charT* p1, const charT* p2)
        basic_regex(const charT* p1, const charT* p2, flag_type f)
        basic_regex(const charT* p, size_t len, flag_type f)
        basic_regex(const string& p)
        basic_regex(const string& p, flag_type f)
    ctypedef basic_regex[char] regex
    ctypedef struct match_results

    bint regex_match[charT](const charT* s, const regex& rgx)
    bint regex_match[charT](const charT* s, const regex& rgx, match_flag_type flags)
    bint regex_match(const const string& s, const regex& rgx)
    bint regex_match(const const string& s, const regex& rgx, match_flag_type flags)

    bint regex_search[charT](const charT* s, const regex& rgx)
    bint regex_search[charT](const charT* s, const regex& rgx, match_flag_type flags)
    bint regex_search(const const string& s, const regex& rgx)
    bint regex_search(const const string& s, const regex& rgx, match_flag_type flags)

    string regex_replace(const string& s, const regex, const string& fmt)
    string regex_replace(const string& s, const regex, const string& fmt, match_flag_type flags)


cdef extern from * nogil:
    """
    #include <boost/regex.hpp>
    #include <iterator>
    #include <string>

    inline bool regex_match(std::string_view s,
                      const boost::regex& rgx,
                      boost::regex_constants::match_flag_type flags = boost::regex_constants::match_default) {
        return boost::regex_match(s.begin(), s.end(), rgx, flags);
    }

    inline bool regex_search(std::string_view s,
                      const boost::regex& rgx,
                      boost::regex_constants::match_flag_type flags = boost::regex_constants::match_default) {
        return boost::regex_search(s.begin(), s.end(), rgx, flags);
    }

    inline std::string regex_replace(std::string_view s,
                      const boost::regex& rgx,
                      const std::string& fmt,
                      boost::regex_constants::match_flag_type flags = boost::regex_constants::match_default) {
        std::string out;
        boost::regex_replace(std::back_inserter(out), s.begin(), s.end(), rgx, fmt, flags);
        return out;
    }
    """

    bint regex_match(const const string_view& s, const regex& rgx)
    bint regex_match(const const string_view& s, const regex& rgx, match_flag_type flags)

    bint regex_search(const const string_view& s, const regex& rgx)
    bint regex_search(const const string_view& s, const regex& rgx, match_flag_type flags)

    string regex_replace(const string_view& s, const regex, const string& fmt)
    string regex_replace(const string_view& s, const regex, const string& fmt, match_flag_type flags)



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

from cython.operator cimport dereference as deref, predecrement as dec, preincrement as inc
from libc.string cimport memmove
from libcpp.string cimport string
from resiliparse_inc.string_view cimport string_view
from resiliparse_inc.cctype cimport isspace, tolower


cdef inline size_t lstrip_c_str(const char** s_ptr, size_t l) nogil:
    """Strip leading white space from a C string."""
    cdef const char* end = deref(s_ptr) + l
    while deref(s_ptr) < end and isspace(deref(s_ptr)[0]):
        inc(deref(s_ptr))
    return end - deref(s_ptr)


cdef inline size_t rstrip_c_str(const char** s_ptr, size_t l) nogil:
    """Strip trailing white space from a C string."""
    cdef const char* end = deref(s_ptr) + l
    while end > deref(s_ptr) and isspace((end - 1)[0]):
        dec(end)
    return end - deref(s_ptr)


cdef inline size_t strip_c_str(const char** s_ptr, size_t l) nogil:
    """Strip leading and trailing white space from a C string."""
    return rstrip_c_str(s_ptr, lstrip_c_str(s_ptr, l))


cdef inline string rstrip_str(string&& s) nogil:
    """Strip trailing white space from a C++ string in-place."""
    cdef const char* start = s.data()
    cdef size_t l = rstrip_c_str(&start, s.size())
    if l != s.size():
        s.resize(l)
    return s


cdef inline string strip_str(string&& s) nogil:
    """Strip leading and trailing white space from a C++ string in-place."""
    cdef const char* start = s.data()
    cdef size_t l = strip_c_str(&start, s.size())
    if l != s.size():
        memmove(s.data(), start, l)
        s.resize(l)
    return s


cdef inline string_view strip_sv(string_view s) nogil:
    """Strip leading and trailing white space from a C++ string_view."""
    cdef const char* start = s.data()
    cdef size_t l = strip_c_str(&start, s.size())
    if start != s.data():
        s.remove_prefix(start - s.data())
    if l != s.size():
        s.remove_suffix(s.size() - l)
    return s


cdef inline string str_to_lower(string s) nogil:
    """Convert a C++ string to lower-case characters."""
    cdef size_t i
    for i in range(s.size()):
        s[i] = tolower(s[i])
    return s



from libc.stdint cimport int64_t
from libcpp.map cimport map
from libcpp.string cimport string
from libcpp.vector cimport vector

from resiliparse_inc.string_view cimport string_view


cdef extern from "<re2/stringpiece.h>" namespace "re2" nogil:
    cdef cppclass StringPiece:
        StringPiece()
        StringPiece(const string_view& str)
        StringPiece(const string& str)
        StringPiece(const char* str)
        StringPiece(const char* str, size_t len)

        const char* begin() const
        const char* end() const

        size_t size() const
        size_t length() const
        bint empty() const
        const char& operator[](size_t i) const
        const char* data() const
        void remove_prefix(size_t n)
        void remove_suffix(size_t n)
        void set(const char * str)
        void set(const char * str, size_t len)

        string as_string() const
        string ToString() const
        void CopyToString(string* target) const
        void AppendToString(string* target) const
        size_t copy(char * buf, size_t n, size_t pos = 0)
        int compare(const StringPiece& x) const
        bint starts_with(const StringPiece& x) const
        bint ends_with(const StringPiece& x) const
        bint ends_with(const StringPiece& x) const

        size_t find(const StringPiece& s, size_t pos = 0) const
        size_t find(char c, size_t pos = 0) const
        size_t rfind(const StringPiece& s) const
        size_t rfind(const StringPiece& s, size_t pos) const
        size_t rfind(char c) const
        size_t rfind(char c, size_t pos) const


cdef extern from "<re2/re2.h>" namespace "re2::Options" nogil:
    cdef enum Encoding:
        EncodingUTF8,
        EncodingLatin1


cdef extern from "<re2/re2.h>" namespace "re2::RE2" nogil:
    cdef enum ErrorCode:
        NoError = 0,
        ErrorInternal,
        ErrorBadEscape,
        ErrorBadCharClass,
        ErrorBadCharRange,
        ErrorMissingBracket,
        ErrorMissingParen,
        ErrorUnexpectedParen,
        ErrorTrailingBackslash,
        ErrorRepeatArgument,
        ErrorRepeatSize,
        ErrorRepeatOp,
        ErrorBadPerlOp,
        ErrorBadUTF8,
        ErrorBadNamedCapture,
        ErrorPatternTooLarge

    enum Anchor:
        UNANCHORED,
        ANCHOR_START,
        ANCHOR_BOTH

    cdef cppclass Options:
        Options()

        Encoding encoding() const
        void set_encoding(Encoding encoding)

        bint posix_syntax() const
        void set_posix_syntax(bint b)

        bint longest_match() const
        void set_longest_match(bint b)

        bint log_errors() const
        void set_log_errors(bint b)

        int64_t max_mem() const
        void set_max_mem(int64_t m)

        bint literal() const
        void set_literal(bint b)

        bint never_nl() const
        void set_never_nl(bint b)

        bint dot_nl() const
        void set_dot_nl(bint b)

        bint never_capture() const
        void set_never_capture(bint b)

        bint case_sensitive() const
        void set_case_sensitive(bint b)

        bint perl_classes() const
        void set_perl_classes(bint b)

        bint word_boundary() const
        void set_word_boundary(bint b)

        bint one_line() const
        void set_one_line(bint b)

        void Copy(const Options& src)
        int ParseFlags() const


cdef extern from "<re2/re2.h>" namespace "re2" nogil:
    cdef cppclass RE2:
        bint ok() const
        const string& pattern() const
        const string& error() const
        ErrorCode error_code() const
        const string& error_arg() const

        int ProgramSize() const
        int ReverseProgramSize() const

        int ProgramFanout(vector[int]* histogram) const
        int ReverseProgramFanout(vector[int]* histogram) const

        PossibleMatchRange(string* min, string* max, int maxlen) const
        int NumberOfCapturingGroups() const
        const map[string, int]& NamedCapturingGroups() const
        const map[int, string]& CapturingGroupNames() const

        bint Match(const StringPiece& text, size_t startpos, size_t endpos, Anchor re_anchor,
                   StringPiece* submatch, int nsubmatch) const
        bint CheckRewriteString(const StringPiece& rewrite, string* error) const
        bint Rewrite(string* out, const StringPiece& rewrite, const StringPiece* vec, int veclen) const


cdef extern from "<re2/re2.h>" namespace "re2::RE2" nogil:
    cdef cppclass Arg

    bint FullMatchN(const StringPiece& text, const RE2& re, const Arg* const args[], int n)
    bint PartialMatchN(const StringPiece& text, const RE2& re, const Arg* const args[], int n)
    bint ConsumeN(StringPiece* input, const RE2& re, const Arg* const args[], int n)
    bint FindAndConsumeN(StringPiece* input, const RE2& re, const Arg* const args[], int n)

    bint FullMatch(const StringPiece& text, const RE2& re)
    bint PartialMatch(const StringPiece& text, const RE2& re)
    bint Consume(StringPiece* input, const RE2& re)
    bint FindAndConsume(StringPiece* input, const RE2& re)

    # bint FullMatch[A](const StringPiece& text, const RE2& re, A&&... a)
    # bint PartialMatch[A](const StringPiece& text, const RE2& re, A&&... a)
    # bint Consume[A](StringPiece* input, const RE2& re, A&&... a)
    # bint FindAndConsume[A](StringPiece* input, const RE2& re, A&&... a)

    bint Replace(string* str, const RE2& re, const StringPiece& rewrite)
    int GlobalReplace(string* str, const RE2& re, const StringPiece& rewrite)
    bint Extract(const StringPiece& text, const RE2& re, const StringPiece& rewrite, string* out)
    string QuoteMeta(const StringPiece& unquoted)
    int MaxSubmatch(const StringPiece& rewrite)


# Stack assignable wrapper (Cython 0.29.x doesn't support cpp_locals yet)
cdef extern from * nogil:
    """
    class RE2Stack {
    public:
        RE2Stack()
            : instance(nullptr) {}
        RE2Stack(const char* pattern)
            : instance(new re2::RE2(pattern)) {}
        RE2Stack(const std::string& pattern)
            : instance(new re2::RE2(pattern)){}
        RE2Stack(const re2::StringPiece& pattern)
            : instance(new re2::RE2(pattern)) {}
        RE2Stack(const re2::StringPiece& pattern, const re2::RE2::Options& options)
            : instance(new re2::RE2(pattern, options)) {}
        RE2Stack(const RE2Stack&) = delete;
        RE2Stack(RE2Stack&&) = delete;
        ~RE2Stack() {
            if (instance) {
                delete instance;
                instance = nullptr;
            }
        }
        RE2Stack& operator=(const RE2Stack&) = delete;
        RE2Stack& operator=(RE2Stack&& other) {
            if (this != &other) {
                instance = other.instance;
                other.instance = nullptr;
            }
            return *this;
        }

        inline const re2::RE2& operator()() const {
            return *instance;
        }

    private:
        re2::RE2* instance;
    };
    """
    cdef cppclass RE2Stack:
        RE2Stack()
        RE2Stack(const char* pattern)
        RE2Stack(const string& pattern)
        RE2Stack(const StringPiece& pattern)
        RE2Stack(const string&, const Options& options)
        RE2Stack(const char* pattern, const Options& options)
        RE2Stack(const StringPiece& pattern, const Options& options)

        inline const RE2& operator()() const



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

from fastwarc.stream_io cimport IOStream, GZipStream, LZ4Stream, FileStream, PythonIOStreamAdapter
from fastwarc.stream_io import StreamError
from fastwarc.warc cimport ArchiveIterator, WarcRecordType


cpdef enum CompressionAlg:
    gzip,
    lz4,
    uncompressed,
    auto


def detect_compression_algorithm(infile_name):
    """
    Try to detect the used compression algorithm from the given filename.

    :param infile_name: input filename
    :return: compression algorithm
    """
    filename = None
    if type(infile_name) is str:
        filename = infile_name
    elif hasattr(infile_name, 'name'):
        filename = infile_name.name

    if type(filename) is str and filename.endswith('.gz'):
        return CompressionAlg.gzip
    elif type(filename) is str and filename.endswith('.lz4'):
        return CompressionAlg.lz4
    elif type(filename) is str and filename.endswith('.warc'):
        return CompressionAlg.uncompressed
    else:
        # Unknown, let stream negotiation try to determine the stream format
        return CompressionAlg.auto


def wrap_warc_stream(warc_in, mode='rb', CompressionAlg comp_alg=auto, **comp_args):
    """
    Wrap WARC stream based on type and desired compression algorithm for use with
    :class:`fastwarc.warc.ArchiveIterator`.

    :param warc_in: input stream or filename
    :param mode: stream open mode
    :param comp_alg: compression algorithm to use
    :param comp_args: optional arguments to pass on to the compressing stream
    :return: wrapped WARC stream
    """
    if type(warc_in) is str:
        stream = FileStream.__new__(FileStream, warc_in, mode)
    elif isinstance(warc_in, IOStream):
        stream = <IOStream>warc_in
    else:
        attr = 'write' if 'w' in mode else 'read'
        if not hasattr(warc_in, attr):
            raise AttributeError(f"Object of type '{type(warc_in).__name__}' has no attribute '{attr}'.")
        stream = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, warc_in)

    if comp_alg == gzip:
        stream = GZipStream.__new__(GZipStream, stream, **comp_args)
    elif comp_alg == lz4:
        stream = LZ4Stream.__new__(LZ4Stream, stream, **comp_args)

    return stream


def recompress_warc_interactive(warc_in, warc_out, CompressionAlg comp_alg_in=auto,
                                CompressionAlg comp_alg_out=auto, **comp_args):
    """
    Recompress WARC file.

    Currently supported compression algorithms: GZip and LZ4.

    :param warc_in: input file name or stream object
    :param warc_out: output file name or stream object
    :param comp_alg_in: compression algorithm of input file
    :param comp_alg_out: compression algorithm to use for output file
    :param comp_args: keyword arguments pass on to the compressor
    :return: generator of records and bytes written in each iteration
    """

    cdef IOStream in_stream
    cdef IOStream out_stream

    if comp_alg_in == auto:
        comp_alg_in = detect_compression_algorithm(warc_in)
    if comp_alg_out == auto:
        if type(warc_in) is str:
            comp_alg_out = detect_compression_algorithm(warc_out)
        else:
            raise StreamError('Illegal output compression algorithm: auto')

    in_stream = wrap_warc_stream(warc_in, 'rb', comp_alg_in)
    out_stream = wrap_warc_stream(warc_out, 'wb', comp_alg_out, **comp_args)

    num = 0
    # noinspection PyTypeChecker
    for record in ArchiveIterator.__new__(ArchiveIterator, in_stream,
                                          parse_http=False, record_types=WarcRecordType.any_type):
        bytes_written = record.write(out_stream)
        yield record, bytes_written

    in_stream.close()
    out_stream.close()


def recompress_warc(warc_in, warc_out, CompressionAlg comp_alg_in=auto, CompressionAlg comp_alg_out=auto, **comp_args):
    """
    Recompress WARC file.

    Currently supported compression algorithms: GZip and LZ4.
    This is a Non-interactive version of :func:`recompress_warc_interactive`.

    :param warc_in: input file name or stream object
    :param warc_out: output file name or stream object
    :param comp_alg_in: compression algorithm of input file
    :param comp_alg_out: compression algorithm to use for output file
    :param comp_args: keyword arguments pass on to the compressor
    :return: number of records recompressed and number of bytes written
    """

    bytes_written_total = 0
    num = 0
    for record, bytes_written in recompress_warc_interactive(warc_in, warc_out, comp_alg_in, comp_alg_out, **comp_args):
        bytes_written_total += bytes_written
        num += 1

    return num, bytes_written


def verify_digests(warc_in, bint verify_payloads=False, CompressionAlg comp_alg=auto):
    """
    Verify block or (optionally) payload digests of all records in a WARC.

    Returns a generator of dicts containing the following structure:
    ```
    {
        "record_id": <ID>,
        "block_digest_ok">: <OK_VAL>
        [ "payload_digest_ok">: <OK_VAL> ]
    }
    ```

    `<OK_VAL>` is either `True` or `False` or `None` if the record has no digest.

    :param warc_in: input WARC file
    :param verify_payloads: verify payload digests
    :param comp_alg: WARC compression algorithm
    :return: generator of dicts containing verification result data
    """

    if comp_alg == auto:
        comp_alg = detect_compression_algorithm(warc_in)

    in_stream = wrap_warc_stream(warc_in, 'rb', comp_alg)

    # noinspection PyTypeChecker
    for record in ArchiveIterator.__new__(ArchiveIterator, in_stream,
                                          parse_http=False, record_types=WarcRecordType.any_type):
        consume = not verify_payloads or not record.is_http
        res = {
            'record_id': record.record_id,
            'block_digest_ok': record.verify_block_digest(consume) if 'WARC-Block-Digest' in record.headers else None
        }

        if verify_payloads:
            if 'WARC-Payload-Digest' in record.headers:
                record.parse_http()
                res['payload_digest_ok'] = record.verify_payload_digest(True)
            else:
                res['payload_digest_ok'] = None

        yield res



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdio cimport FILE
from libcpp.string cimport string

from resiliparse_inc.zlib cimport *
from resiliparse_inc.lz4hc cimport *
from resiliparse_inc.lz4frame cimport *
from resiliparse_inc.string_view cimport string_view


cdef class IOStream:
    cdef size_t read_(self, char* out, size_t size) except -1
    cdef size_t write_(self, const char* data, size_t size) except -1
    cpdef size_t tell(self) except -1
    cpdef void seek(self, size_t offset) except *
    cpdef void flush(self)  except *
    cpdef void close(self) except *


cdef class PythonIOStreamAdapter(IOStream):
    cdef object py_stream


cdef class BytesIOStream(IOStream):
    cdef string buffer
    cdef size_t pos

    cpdef string getvalue(self)


cdef class FileStream(IOStream):
    cdef FILE* fp

    cdef void open_(self, const char* path, char* mode=*) except *


cdef class CompressingStream(IOStream):
    cpdef size_t begin_member(self)
    cpdef size_t end_member(self)


cdef extern from * nogil:
    """
    /* Helper enum for keeping track of current compressing stream state. */
    enum CompressingStreamState {
        UNINIT = 0,
        DECOMPRESSING = 1,
        COMPRESSING = 2
    };
    """
    ctypedef enum CompressingStreamState:
        UNINIT,
        DECOMPRESSING,
        COMPRESSING


cdef class GZipStream(CompressingStream):
    cdef IOStream raw_stream
    cdef string working_buf
    cdef unsigned int working_buf_filled
    cdef z_stream zst
    cdef int window_bits
    cdef char stream_state
    cdef size_t stream_pos
    cdef bint member_started
    cdef int compression_level

    cpdef void prepopulate(self, bint deflate, const string& initial_data)
    cdef void _init_z_stream(self, bint zlib) nogil
    cdef void _free_z_stream(self) nogil
    cdef void _reinit_z_stream(self, bint deflate) nogil
    cdef bint _refill_working_buf(self, size_t size) except -1


cdef class LZ4Stream(CompressingStream):
    cdef IOStream raw_stream
    cdef LZ4F_cctx* cctx
    cdef LZ4F_dctx* dctx
    cdef LZ4F_preferences_t prefs
    cdef string working_buf
    cdef size_t working_buf_read
    cdef size_t working_buf_filled
    cdef bint frame_started
    cdef size_t stream_pos

    cpdef void prepopulate(self, const string& initial_data)
    cdef void _free_ctx(self) nogil


cdef class BrotliStream(CompressingStream):
    cdef IOStream raw_stream
    cdef size_t quality
    cdef size_t lgwin
    cdef size_t lgblock
    cdef object compressor
    cdef string working_buf
    cdef CompressingStreamState stream_state


cdef class BufferedReader:
    cdef IOStream stream
    cdef string errstr
    cdef string buf
    cdef string_view buf_view
    cdef string_view limited_buf_view
    cdef size_t stream_pos
    cdef size_t limit
    cdef size_t limit_consumed
    cdef bint negotiate_stream
    cdef bint stream_started
    cdef bint stream_is_compressed

    cdef inline void set_limit(self, size_t offset) nogil
    cdef inline void reset_limit(self) nogil
    cdef bint detect_stream_type(self) except 0

    cpdef string read(self, size_t size=*) except *
    cpdef string readline(self, bint crlf=*, size_t max_line_len=*)  except *
    cpdef size_t tell(self) except -1
    cpdef size_t consume(self, size_t size=*) except -1
    cpdef void close(self) except *

    cdef bint _fill_buf(self) except -1
    cdef inline string_view* _get_buf(self) nogil
    cdef inline size_t _consume_buf(self, size_t size) nogil



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

cimport cython
from cython.operator cimport dereference as deref, preincrement as inc
from libc.stdint cimport uint16_t
from libc.string cimport memchr
from libcpp.string cimport npos as strnpos, string, to_string
from libcpp.vector cimport vector

import binascii
import codecs
import base64
import datetime
from email.utils import parsedate_to_datetime
import hashlib
import typing as t
import uuid
import warnings

from resiliparse_common.string_util cimport str_to_lower, strip_str, strip_c_str
from resiliparse_inc.cctype cimport isspace
from resiliparse_inc.cstdlib cimport strtol
from resiliparse_inc.utility cimport move

from fastwarc.stream_io cimport BufferedReader, BytesIOStream, CompressingStream, GZipStream, BrotliStream, \
    IOStream, PythonIOStreamAdapter
from fastwarc.stream_io import ReaderStaleError


cdef const char* _enum_record_type_to_str(WarcRecordType record_type) nogil:
    if record_type == warcinfo:
        return b'warcinfo'
    elif record_type == response:
        return b'response'
    elif record_type == resource:
        return b'resource'
    elif record_type == request:
        return b'request'
    elif record_type == metadata:
        return b'metadata'
    elif record_type == revisit:
        return b'revisit'
    elif record_type == conversion:
        return b'conversion'
    elif record_type == continuation:
        return b'continuation'
    else:
        return b'unknown'


cdef WarcRecordType _str_record_type_to_enum(const string& record_type) nogil:
    cdef string record_type_lower = str_to_lower(record_type)
    if record_type_lower == b'warcinfo':
        return warcinfo
    elif record_type_lower == b'response':
        return response
    elif record_type_lower == b'resource':
        return resource
    elif record_type_lower == b'request':
        return request
    elif record_type_lower == b'metadata':
        return metadata
    elif record_type_lower == b'revisit':
        return revisit
    elif record_type_lower == b'conversion':
        return conversion
    elif record_type_lower == b'continuation':
        return continuation
    else:
        return unknown


class CaseInsensitiveStr(str):
    """Case-insensitive str implementation for use as dict key."""

    def __hash__(self):
        return hash(self.casefold())

    def __eq__(self, other):
        return isinstance(other, str) and self.casefold() == other.casefold()


class CaseInsensitiveStrDict(dict):
    """Case-insensitive str dict."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._header_obj = None

    def __getitem__(self, key not None):
        return super().__getitem__(CaseInsensitiveStr(key))

    def __setitem__(self, key not None, value not None):
        super().__setitem__(CaseInsensitiveStr(key), value)
        if hasattr(self, '_header_obj') and self._header_obj is not None:
            self._header_obj[key] = value

    def __contains__(self, key not None):
        return super().__contains__(CaseInsensitiveStr(key))

    def get(self, key not None, value=None):
        return super().get(CaseInsensitiveStr(key), value)

    def setdefault(self, key not None, value=None):
        return super().setdefault(CaseInsensitiveStr(key), value)

    def pop(self, key not None):
        return super().pop(CaseInsensitiveStr(key))

    def update(self, it=None, **kwargs):
        if hasattr(it, 'items'):
            it = it.items()
        if it is not None:
            for k, v in it:
                self.__setitem__(k, v)
        for k, v in kwargs.items():
            self.__setitem__(k, v)

    def clear(self) -> None:
        if self._header_obj is not None:
            self._header_obj.clear()
        else:
            super().clear()


def _rebuild_warc_header_map(encoding, status_line, headers):
    """Unpickle handler for :class:`WarcHeaderMap`."""
    cdef WarcHeaderMap header_map = WarcHeaderMap.__new__(WarcHeaderMap, encoding)
    header_map._status_line = status_line
    for k, v in headers:
        header_map.append_header(k, v)
    return header_map


# noinspection PyAttributeOutsideInit
cdef class WarcHeaderMap:
    """
    __init__(self, encoding='utf-8')

    Dict-like type representing a WARC or HTTP header block.

    :param encoding: header source encoding
    :type encoding: str
    """

    def __init__(self, *args, **kwargs):
        pass

    def __reduce__(self):
        return _rebuild_warc_header_map, (self._enc, self._status_line, self._headers)

    def __cinit__(self, str encoding='utf-8'):
        self._enc = encoding
        self._dict_cache = None
        self._dict_cache_stale = True

    def __getitem__(self, header_key):
        return self.asdict()[header_key]

    def __setitem__(self, header_key, header_value):
        header_key = header_key.strip()
        header_value = header_value.replace('\r\n', ' ').replace('\n', ' ').strip()
        self.set_header(header_key.encode(self._enc, errors='ignore'),
                        header_value.encode(self._enc, errors='ignore'))

    def __iter__(self):
        """
        __iter__(self)

        Iterate all header map items.

        :rtype: t.Iterable[(str, str)]
        """
        cdef str_pair h
        yield from ((CaseInsensitiveStr(h[0].decode(self._enc, errors='ignore')),
                     h[1].decode(self._enc, errors='ignore')) for h in self._headers)

    def __repr__(self):
        return repr(self.astuples())

    def __str__(self):
        return str(self.astuples())

    def __len__(self):
        return len(self.astuples())

    def __contains__(self, item):
        return item in self.asdict()

    # noinspection PyProtectedMember
    def __eq__(self, other):
        if not isinstance(other, WarcHeaderMap):
            return False
        if self._status_line != (<WarcHeaderMap>other)._status_line:
            return False
        for (k1, v1), (k2, v2) in zip(self.__iter__(), other.__iter__()):
            if k1 != k2 or v1 != v2:
                return False
        return True

    @property
    def status_line(self) -> str:
        """
        Header status line.

        :type: str
        """
        return self._status_line.decode(self._enc, errors='ignore')

    @status_line.setter
    def status_line(self, status_line not None):
        """
        Set status line contents.

        :param status_line: new status line
        :type status_line: str
        """
        self._status_line = status_line.encode(self._enc, errors='ignore')

    @property
    def status_code(self) -> t.Optional[int]:
        """
        HTTP status code (unset if header block is not an HTTP header block).

        :type: int or None
        """
        if self._status_line.find(<char*>b'HTTP/') != 0:
            return None
        s = self._status_line.split(b' ', 2)
        if len(s) < 2 or not s[1].isdigit():
            return None
        return int(s[1])

    @property
    def reason_phrase(self) -> t.Optional[str]:
        """
        HTTP reason phrase (unset if header block is not an HTTP header block).

        :type: str or None
        """
        if self._status_line.find(<char*>b'HTTP/') != 0:
            return None
        s = self._status_line.split(b' ', 3)
        if len(s) < 3 or not s[1].isdigit():
            return None
        return s[2].decode(self._enc, errors='ignore')

    def append(self, key not None, value not None):
        """
        append(self, key, value)

        Append header (use if header name is not unique).

        :param key: header key
        :type key: str
        :param value: header value
        :type value: str
        """
        key = key.strip()
        value = value.replace('\r\n', ' ').replace('\n', ' ').strip()
        self.append_header(key.encode(self._enc), value.encode(self._enc))

    def get(self, key not None, default=None) -> str:
        """
        get(self, key, default=None)

        Get header value or ``default``.

        If multiple headers have the same key, only the last occurrence will be returned.

        :param key: header key
        :type key: str
        :param default: default value if ``key`` not found
        :type default: str
        :rtype: str
        """
        return self.asdict().get(key, default)

    def items(self):
        """
        items(self)

        Item view of keys and values.

        If multiple headers have the same key, only the last occurrence will be returned.

        :rtype: t.Iterable[(str, str)]
        """
        return self.asdict().items()

    def keys(self):
        """
        keys(self)

        Iterable of header keys.

        If multiple headers have the same key, only the last occurrence will be returned.

        :rtype: t.Iterable[str]
        """
        return self.asdict().keys()

    def values(self):
        """
        values(self)

        Iterable of header values.

        If multiple headers have the same key, only the last occurrence will be returned.

        :rtype: t.Iterable[str]
        """
        return self.asdict().values()

    def asdict(self) -> t.Dict[str, str]:
        """
        asdict(self)

        Headers as Python dict.

        If multiple headers have the same key, only the last occurrence will be returned.

        :rtype: t.Dict[str, str]
        """
        cdef str_pair h
        if self._dict_cache_stale:
            if self._dict_cache is None:
                self._dict_cache = <dict>CaseInsensitiveStrDict()
            self._dict_cache._header_obj = None
            self._dict_cache.clear()
            self._dict_cache.update({
                h[0].decode(self._enc, errors='ignore'): h[1].decode(self._enc, errors='ignore')
                for h in self._headers})
            self._dict_cache_stale = False
            self._dict_cache._header_obj = self
        return self._dict_cache

    def astuples(self) -> t.Tuple[t.Tuple[str, str]]:
        """
        astuples(self)

        Headers as a series of tuples, including multiple headers with the same key.
        Use this over :meth:`asdict` if header keys are not necessarily unique.

        :rtype: ((str, str),)
        """
        return tuple(self.__iter__())

    cpdef void clear(self):
        """
        clear(self)

        Clear all headers.
        """
        if self._dict_cache is not None:
            self._dict_cache._header_obj = None
            self._dict_cache.clear()
            self._dict_cache._header_obj = self
        self._headers.clear()
        self._status_line.clear()
        self._dict_cache_stale = True

    cpdef size_t write(self, IOStream stream) except -1:
        """Write header block into stream."""
        cdef size_t bytes_written = 0
        if not self._status_line.empty():
            bytes_written += stream.write_(self._status_line.data(), self._status_line.size())
            bytes_written += stream.write_(<char*>b'\r\n', 2)

        cdef vector[str_pair].iterator it = self._headers.begin()
        while it != self._headers.end():
            if not deref(it)[0].empty():
                bytes_written += stream.write_(deref(it)[0].data(), deref(it)[0].size())
                bytes_written += stream.write_(<char*>b': ', 2)
            bytes_written += stream.write_(deref(it)[1].data(), deref(it)[1].size())
            bytes_written += stream.write_(<char*>b'\r\n', 2)
            inc(it)
        return bytes_written

    cdef inline void set_status_line(self, const string& status_line):
        """Set status line string."""
        self._status_line = status_line

    cdef string find_header(self, const string& header_key, const string& default):
        """Return value of first header occurrence or ``default`` (linear complexity)."""
        cdef string header_key_lower = str_to_lower(header_key)
        cdef vector[str_pair].iterator it = self._headers.begin()
        while it != self._headers.end():
            if str_to_lower(deref(it)[0]) == header_key_lower:
                return deref(it)[1]
            inc(it)

        return default

    cdef void set_header(self, const string& header_key, const string& header_value):
        """Set new header or overwrite existing header if it exists."""
        self._dict_cache_stale = True
        cdef string header_key_lower = str_to_lower(header_key)
        cdef vector[str_pair].iterator it = self._headers.begin()
        while it != self._headers.end():
            if str_to_lower(deref(it)[0]) == header_key_lower:
                deref(it)[1] = header_value
                return
            inc(it)

        self._headers.push_back((header_key, header_value))

    cdef inline void append_header(self, const string& header_key, const string& header_value):
        """Append header (use if header key is not unique)."""
        self._headers.push_back((header_key, header_value))
        self._dict_cache_stale = True

    cdef void add_continuation(self, const string& header_continuation_value):
        """Append value to previous header."""
        if not self._headers.empty():
            self._headers.back()[1].append(<char*>b' ')
            self._headers.back()[1].append(header_continuation_value)
        else:
            # This should no happen, but what can we do?!
            self._headers.push_back((<char*>b'', header_continuation_value))
        self._dict_cache_stale = True


def _rebuild_warc_record(record_type, headers, is_http, http_parsed, http_charset, http_headers,
                         content_length, content_bytes, stream_pos):
    """Unpickle handler for :class:`WarcRecord`."""
    cdef WarcRecord record = WarcRecord.__new__(WarcRecord)
    record._record_type = record_type
    record._headers = headers
    record._is_http = is_http
    record._http_parsed = http_parsed
    record._http_charset = http_charset
    record._http_headers = http_headers
    record._content_length = content_length
    record._reader = BufferedReader.__new__(BufferedReader, BytesIOStream.__new__(BytesIOStream, content_bytes))
    record._stream_pos = stream_pos
    return record


# noinspection PyProtectedMember, PyAttributeOutsideInit
cdef class WarcRecord:
    """
    __init__(self)

    A WARC record.

    WARC records are picklable, but pickling will :meth:`freeze()` the WARC record.
    """

    def __cinit__(self):
        self._record_type = unknown
        self._is_http = False
        self._http_parsed = False
        self._content_length = 0
        self._headers = WarcHeaderMap.__new__(WarcHeaderMap, 'utf-8')
        self._http_headers = None   # type: WarcHeaderMap | None
        self._stream_pos = 0
        self._stale = False
        self._frozen = False

    def __reduce__(self):
        self.freeze()
        return _rebuild_warc_record, (
            self._record_type,
            self._headers,
            self._is_http,
            self._http_parsed,
            self._http_charset,
            self._http_headers,
            self._content_length,
            (<BytesIOStream>self._reader.stream).buffer,
            self._stream_pos
        )

    cdef inline bint _assert_not_stale(self) except 0:
        if self._stale:
            raise ReaderStaleError('Reader is stale. Use freeze() if you want to retain record payload.')
        return True

    @property
    def record_id(self) -> str:
        """
        Record ID (same as ``headers['WARC-Record-ID']``).

        :type: str
        """
        return self._headers['WARC-Record-ID']

    @property
    def record_type(self) -> WarcRecordType:
        """
        Record type (same as ``headers['WARC-Type']``.

        :type: WarcRecordType
        """
        return self._record_type

    @record_type.setter
    def record_type(self, WarcRecordType record_type):
        """
        record_type(self, record_type)

        Set record type.

        :param record_type: record type
        :type record_type: WarcRecordType
        """
        self._record_type = record_type
        self._headers.set_header(<char*>b'WARC-Type', _enum_record_type_to_str(record_type))

    @property
    def record_date(self):
        """
        WARC Date.

        :type: datetime.datetime | None
        """
        try:
            date = self._headers.get('WARC-Date', '').replace('Z', '+00:00')
            return datetime.datetime.fromisoformat(date)
        except ValueError:
            return None

    @record_date.setter
    def record_date(self, date):
        """
        record_date(self, date)

        Set WARC Date.

        :param date: datetime object
        :type: datetime.datetime | None
        """
        if not isinstance(date, datetime.datetime):
            raise TypeError('Invalid type: ' + type(date))

        if date.tzinfo is None:
            raise ValueError('Trying to set naive datetime without timezone info.')

        self._headers.set_header(b'WARC-Date', date.isoformat().replace('+00:00', 'Z').encode())

    @property
    def headers(self) -> WarcHeaderMap:
        """
        WARC record headers.

        :type: WarcHeaderMap
        """
        return self._headers

    @property
    def is_http(self) -> bool:
        """
        Whether record is an HTTP record.

        Modifying this property will also affect the ``Content-Type`` of this record.

        :type: bool
        """
        return self._is_http

    @is_http.setter
    def is_http(self, bint is_http):
        self._is_http = is_http
        if self._is_http:
            if self._record_type == WarcRecordType.request:
                self._headers.set_header(<char *> b'Content-Type', <char *> b'application/http; msgtype=request')
            elif self._record_type == WarcRecordType.response:
                self._headers.set_header(<char*>b'Content-Type', <char*>b'application/http; msgtype=response')
            else:
                self._headers.set_header(<char*>b'Content-Type', <char*>b'application/http')

    @property
    def is_http_parsed(self) -> bool:
        """
        Whether HTTP headers have been parsed.

        :type: bool
        """
        return self._http_parsed

    @property
    def http_headers(self) -> t.Optional[WarcHeaderMap]:
        """
        HTTP headers if record is an HTTP record and HTTP headers have been parsed yet.

        :type: WarcHeaderMap or None
        """
        return self._http_headers

    @property
    def http_content_type(self) -> t.Optional[str]:
        """
        Plain HTTP Content-Type without additional fields such as ``charset=``.

        :type: str or None
        """
        if not self._http_parsed:
            return None

        cdef string content_type = self._http_headers.find_header(<char*>b'content-type', <char*>b'')
        if content_type.empty():
            return None

        content_type = strip_str(move(content_type.substr(0, content_type.find(<char*>b';'))))
        return content_type.decode(self._http_headers._enc, errors='ignore')

    @property
    def http_charset(self) -> t.Optional[str]:
        """
        HTTP charset/encoding as returned by the server or ``None`` if no valid charset is set.
        A returned string is guaranteed to be a valid Python encoding name.

        :type: str or None
        """
        if not self._http_parsed or self._http_charset == <char*>b'_':
            return None

        if not self._http_charset.empty():
            return self._http_charset.decode(errors='ignore')

        cdef string content_type = self._http_headers.find_header(<char*>b'content-type', <char*>b'')
        cdef size_t pos = content_type.find(<char*>b';')
        if pos == strnpos:
            return None
        pos = content_type.find(<char*>b'charset=', pos)
        if pos == strnpos:
            return None

        pos += 8
        cdef size_t pos_end = content_type.find(<char*>b';', pos)
        if pos_end != strnpos:
            pos_end = pos_end - pos

        self._http_charset = str_to_lower(strip_str(move(content_type.substr(pos, pos_end))))
        try:
            codecs.lookup(self._http_charset.decode(errors='ignore'))
        except LookupError:
            self._http_charset = <char*>b'_'
            return None

        return self._http_charset.decode(errors='ignore')

    @property
    def http_date(self):
        """
        Parsed HTTP ``Date`` header or ``None`` if server did not return a valid HTTP date.

        :type: datetime.datetime | None
        """
        if not self._http_parsed:
            return None

        try:
            return parsedate_to_datetime(self._http_headers.get('Date', ''))
        except (ValueError, TypeError):
            return None

    @property
    def http_last_modified(self):
        """
        Parsed HTTP ``Last-Modified`` header or ``None`` if server did not return a valid HTTP modification date.

        :type: datetime.datetime | None
        """
        if not self._http_parsed:
            return None

        try:
            return parsedate_to_datetime(self._http_headers.get('Last-Modified', ''))
        except (ValueError, TypeError):
            return None

    @property
    def content_length(self) -> int:
        """
        Remaining WARC record length in bytes (not necessarily the same as the ``Content-Length`` header).

        :type: int
        """
        return self._content_length

    @property
    def reader(self) -> BufferedReader:
        """
        Reader for the remaining WARC record content.

        :type: BufferedReader
        """
        self._assert_not_stale()
        return self._reader

    @property
    def stream_pos(self) -> int:
        """
        WARC record start offset in the original (uncompressed) stream.

        :type: int
        """
        return self._stream_pos

    cpdef void init_headers(self, size_t content_length=0, WarcRecordType record_type=no_type, bytes record_urn=None):
        """
        init_headers(self, content_length=0, record_type=no_type, record_urn=None)

        Initialize mandatory headers in a fresh :class:`WarcRecord` instance.

        :param content_length: WARC record body length in bytes
        :type content_length: int
        :param record_type: WARC-Type
        :type record_type: WarcRecordType
        :param record_urn: WARC-Record-ID as URN without ``'<'``, ``'>'`` (if unset, a random URN will be generated)
        :type record_urn: bytes
        """
        if record_urn is None:
            record_urn = uuid.uuid4().urn.encode()

        if record_type == no_type:
            record_type = self.record_type
        if record_type == any_type or record_type == no_type:
            record_type = unknown
        self._record_type = record_type

        self._headers.clear()
        self._headers.set_status_line(<char*>b'WARC/1.1')
        self._headers.append_header(<char*>b'WARC-Type', _enum_record_type_to_str(record_type))
        self._headers.append_header(<char*>b'WARC-Date', datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ').encode())
        self._headers.append_header(<char*>b'WARC-Record-ID', b''.join((b'<', record_urn, b'>')))
        self._headers.append_header(<char*>b'Content-Length', to_string(<long int>content_length))

    cpdef void set_bytes_content(self, bytes b):
        """
        set_bytes_content(self, b)

        Set WARC body.

        :param b: body as bytes
        :type b: bytes
        """
        self._reader = BufferedReader.__new__(BufferedReader, BytesIOStream(b))
        self._content_length = len(b)
        self._headers.set_header(<char*>b'Content-Length', to_string(<long int>self._content_length))
        self._stale = False

    cpdef bint parse_http(self, bint strict_mode=True, str auto_decode='none') except 0:
        """
        parse_http(self, strict_mode=True, auto_decode='none')

        Parse HTTP headers and advance content reader.

        It is safe to call this method multiple times, even if the record is not an HTTP record.

        :param strict_mode: enforce ``CRLF`` line endings, setting this to ``False`` will allow plain ``LF`` also
        :type: strict_mode: bool
        :param auto_decode: automatically decode record body if Content-Encoding or Transfer-Encoding headers are set
                            (accepted values: ``'none'`` [default], ``'content'``, ``'transfer'``, ``'all'``)
        :type auto_decode: str
        """
        self._assert_not_stale()
        if self._http_parsed or not self._is_http:
            return True
        self._http_headers = WarcHeaderMap.__new__(WarcHeaderMap, 'iso-8859-15')
        cdef size_t num_bytes = parse_header_block(self.reader, self._http_headers, True, strict_mode)
        self._content_length = self._content_length - num_bytes
        self._http_parsed = True

        if auto_decode in ['all', 'content', 'transfer']:
            self._init_content_decoder(auto_decode)

        return True

    cdef bint _init_content_decoder(self, str decode_mode) except -1:
        """
        Init content decoder based on the Content-Encoding and Transfer-Encoding headers.
        """
        cdef size_t delim_pos
        cdef size_t prev_delim_pos
        cdef string enc_str
        cdef string enc_name
        cdef vector[string] encodings
        cdef size_t i
        for i in range(2):
            if i == 0 and decode_mode in ['all', 'content']:
                enc_str = self._http_headers.find_header(b'Content-Encoding', b'')
            elif i == 1 and decode_mode in ['all', 'transfer']:
                enc_str = self._http_headers.find_header(b'Transfer-Encoding', b'')
            else:
                continue

            delim_pos = 0
            prev_delim_pos = 0
            while not enc_str.empty() and delim_pos != strnpos:
                delim_pos = enc_str.find(b',', prev_delim_pos)
                if delim_pos != strnpos:
                    enc_name = strip_str(move(enc_str.substr(prev_delim_pos, delim_pos - prev_delim_pos)))
                else:
                    enc_name = strip_str(move(enc_str.substr(prev_delim_pos)))
                if not enc_name.empty():
                    encodings.push_back(enc_name)
                enc_name.clear()
                prev_delim_pos = delim_pos + 1

        if encodings.empty():
            return True

        self.freeze()
        for i in reversed(range(0, encodings.size())):
            if encodings[i] == b'gzip' or encodings[i] == b'x-gzip':
                self._reader.stream = GZipStream(self._reader.stream)
            elif encodings[i] == b'deflate':
                self._reader.stream = GZipStream(self._reader.stream, zlib=True)
            elif encodings[i] == b'br':
                self._reader.stream = BrotliStream(self._reader.stream)
            elif encodings[i] == b'chunked':
                warnings.warn('Chunked encoding not supported, use Resiliparse for decoding payload.')
                return False
            elif encodings[i] == b'identity' or encodings[i] == b'none':
                # Non-standard in Content-Encoding
                continue
            else:
                warnings.warn('Unsupported encoding: ' + encodings[i].decode())
                return False

        return True

    # noinspection PyTypeChecker
    cpdef size_t write(self, stream, bint checksum_data=False, bytes payload_digest=None,
                       size_t chunk_size=16384) except -1:
        """
        write(self, stream, checksum_data=False, chunk_size=16384)

        Write WARC record onto a stream.

        :param stream: output stream
        :param checksum_data: add block and payload digest headers
        :type checksum_data: bool
        :param payload_digest: optional SHA-1 payload digest as bytes
        :type payload_digest: bytes
        :param chunk_size: write block size
        :type chunk_size: int
        :return: number of bytes written
        :rtype: int
        """
        self._assert_not_stale()

        # If the raw byte content hasn't been parsed, we can simply pass it through
        if not checksum_data and not self._http_parsed:
            return self._write_impl(self.reader, stream, True, chunk_size)

        # Otherwise read everything into memory for content-length correction and checksumming
        cdef BytesIOStream block_buf = BytesIOStream()

        block_hash = hashlib.sha1() if checksum_data else None
        payload_hash = hashlib.sha1() if checksum_data and not payload_digest and self._http_parsed else None

        if self._http_parsed:
            self._http_headers.write(block_buf)
            block_buf.write_(<char*>b'\r\n', 2)

            if checksum_data:
                block_hash.update(block_buf.getvalue())

        cdef string payload_data
        while True:
            payload_data = self.reader.read(chunk_size)
            if payload_data.empty():
                break

            if checksum_data:
                block_hash.update(payload_data.data()[:payload_data.size()])
                if payload_hash is not None:
                    payload_hash.update(payload_data.data()[:payload_data.size()])
            block_buf.write_(payload_data.data(), payload_data.size())

        self._headers.set_header(<char*>b'Content-Length', to_string(<long int>block_buf.tell()))
        if payload_hash:
            payload_digest = payload_hash.digest()
        if payload_digest is not None:
            self._headers.set_header(<char*>b'WARC-Payload-Digest', b'sha1:' + base64.b32encode(payload_digest))
        if checksum_data:
            self._headers.set_header(<char*>b'WARC-Block-Digest', b'sha1:' + base64.b32encode(block_hash.digest()))

        block_buf.seek(0)
        return self._write_impl(block_buf, stream, False, chunk_size)

    cdef size_t _write_impl(self, in_stream, out_stream, bint write_payload_headers, size_t chunk_size) except -1:
        cdef IOStream out_stream_wrapped
        cdef size_t bytes_written = 0
        cdef bint compress_member_started = False

        if isinstance(out_stream, IOStream):
            out_stream_wrapped = <IOStream>out_stream

            if isinstance(out_stream, CompressingStream):
                bytes_written = (<CompressingStream>out_stream_wrapped).begin_member()
                compress_member_started = True

        elif isinstance(out_stream, object) and hasattr(out_stream, 'write'):
            out_stream_wrapped = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, out_stream)
        else:
            raise TypeError(f"Object of type '{type(out_stream).__name__}' is not a valid stream.")

        cdef BufferedReader in_reader_wrapped = None
        cdef IOStream in_stream_wrapped = None
        if isinstance(in_stream, BufferedReader):
            in_reader_wrapped = <BufferedReader>in_stream
        elif isinstance(in_stream, IOStream):
            in_stream_wrapped = <IOStream>in_stream
        elif isinstance(in_stream, object) and hasattr(in_stream, 'read'):
            in_stream_wrapped = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, in_stream)
        else:
            raise TypeError(f"Object of type '{type(in_stream).__name__}' is not a valid stream.")

        bytes_written += self._headers.write(out_stream_wrapped)
        bytes_written += out_stream_wrapped.write_(b'\r\n', 2)

        if write_payload_headers and self._http_parsed:
            bytes_written += self._http_headers.write(out_stream_wrapped)
            bytes_written += out_stream_wrapped.write_(b'\r\n', 2)

        cdef string buffer
        cdef size_t bytes_read = 0
        while True:
            if in_reader_wrapped is not None:
                buffer = in_reader_wrapped.read(chunk_size)
                bytes_read = buffer.size()
            else:
                if buffer.size() < chunk_size:
                    buffer.resize(chunk_size)
                bytes_read = in_stream_wrapped.read_(buffer.data(), chunk_size)
            if bytes_read == 0:
                break
            bytes_written += out_stream_wrapped.write_(buffer.data(), bytes_read)

        bytes_written += out_stream_wrapped.write_(b'\r\n\r\n', 4)

        if compress_member_started:
            bytes_written += (<CompressingStream>out_stream_wrapped).end_member()

        return bytes_written

    cdef bint _verify_digest(self, const string& digest_str, bint consume) except -1:
        self._assert_not_stale()

        cdef size_t sep_pos = digest_str.find(b':')
        if sep_pos == strnpos:
            return False

        cdef string alg = digest_str.substr(0, sep_pos)
        if alg in (b'sha1', b'md5', b'sha256', b'sha512'):
            h = getattr(hashlib, alg.decode())()
        else:
            warnings.warn(f'Unsupported hash algorithm: "{alg.decode()}".')
            return False

        cdef bytes digest_bytes
        try:
            digest_bytes = base64.b32decode(digest_str.substr(sep_pos + 1))
        except binascii.Error:
            try:
                # Hex digests are non-standard, but are created by some libraries such as warcprox
                digest_bytes = binascii.unhexlify(digest_str.substr(sep_pos + 1))
            except binascii.Error:
                warnings.warn('Invalid digest encoding.')
                return False

        if not consume and not self._frozen:
            self.freeze()

        cdef string block
        while True:
            block = self._reader.read(4096)
            if block.empty():
                break
            h.update(block)

        if not consume:
            self._reader.stream.seek(0)

        return h.digest() == digest_bytes

    cpdef bint freeze(self) except 0:
        """
        freeze(self)

        "Freeze" a record by baking in the remaining payload stream contents.

        Freezing a record makes the :class:`WarcRecord` instance copyable and reusable by decoupling
        it from the underlying raw WARC stream. Instead of reading directly from the raw stream, a
        frozen record maintains an internal buffer the size of the remaining payload stream contents
        at the time of calling ``freeze()``.

        Freezing a record will advance the underlying raw stream.
        """
        self._assert_not_stale()

        cdef string buffer = self._reader.read()
        cdef BytesIOStream bio = BytesIOStream.__new__(BytesIOStream)
        bio.buffer = move(buffer)
        self._reader = BufferedReader.__new__(BufferedReader, bio)
        self._frozen = True
        return True

    cpdef bint verify_block_digest(self, bint consume=False) except -1:
        """
        verify_block_digest(self, consume=False)

        Verify whether record block digest is valid.

        :param consume: do not create an in-memory copy of the record stream
                        (will fully consume the rest of the record)
        :type consume: bool
        :return: ``True`` if digest exists and is valid
        :rtype: bool
        """
        return self._verify_digest(self._headers.find_header(<char*>b'WARC-Block-Digest', <char*>b''), consume)

    cpdef bint verify_payload_digest(self, bint consume=False) except -1:
        """
        verify_payload_digest(self, consume=False)

        Verify whether record payload digest is valid.

        :param consume: do not create an in-memory copy of the record stream
                        (will fully consume the rest of the record)
        :type consume: bool
        :return: ``True`` if record is HTTP record and digest exists and is valid
        :rtype: bool
        """
        if not self._http_parsed:
            return False
        return self._verify_digest(self._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b''), consume)


# noinspection PyProtectedMember
cdef size_t parse_header_block(BufferedReader reader, WarcHeaderMap target, bint has_status_line,
                               bint strict_mode=True) except -1:
    """
    parse_header_block(reader, target, has_status_line=False, strict_mode=True)

    Helper function for parsing WARC or HTTP header blocks.

    :param reader: input reader
    :type reader: BufferedReader
    :param target: header map to fill
    :type reader: WarcHeaderMap
    :param has_status_line: whether first line is a status line or already a header
    :type has_status_line: bool
    :param strict_mode: enforce ``CRLF`` line endings, setting this to ``False`` will allow plain ``LF`` also
    :type: strict_mode: bool
    :return: number of bytes read from `reader`
    :rtype: int
    """
    cdef string line
    cdef string header_key, header_value
    cdef const char* delim_ptr
    cdef size_t bytes_consumed = 0

    # Temporary variables for white-space stripping
    cdef const char* str_start
    cdef size_t str_len

    while True:
        line = reader.readline(strict_mode)
        bytes_consumed += line.size()

        if line == b'\r\n' or line == b'' or (not strict_mode and line == b'\n'):
            break

        if isspace(line[0]):
            # Continuation line
            target.add_continuation(strip_str(line.substr()))
            continue

        if has_status_line:
            target.set_status_line(strip_str(line.substr()))
            has_status_line = False
            continue

        str_start = line.data()
        delim_ptr = <const char*>memchr(str_start, <int>b':', line.size())
        if delim_ptr == NULL:
            # Invalid header, try to preserve it as best we can
            target.add_continuation(strip_str(line.substr()))
        else:
            str_len = strip_c_str(&str_start, delim_ptr - str_start)
            header_key = string(str_start, str_len)

            if delim_ptr == &line.back():
                header_value = string()
            else:
                inc(delim_ptr)
                str_len = strip_c_str(&delim_ptr, &line.back() - delim_ptr)
                header_value = string(delim_ptr, str_len)
            target.append_header(move(header_key), move(header_value))

    return bytes_consumed


# noinspection PyProtectedMember, PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class ArchiveIterator:
    """
    __init__(self, stream, record_types=any_type, parse_http=True, min_content_length=-1, max_content_length=-1, \
             func_filter=None, verify_digests=False, strict_mode=True, auto_decode='none')

    WARC record stream iterator.

    :param stream: input stream (preferably an :class:`~fastwarc.stream_io.IOStream`,
                   but any file-like Python object is fine)
    :param parse_http: whether to parse HTTP records automatically (disable for better performance if not needed)
    :type parse_http: bool
    :param record_types: bitmask of :class:`WarcRecordType` record types to return (others will be skipped)
    :type record_types: int
    :param min_content_length: skip records with Content-Length less than this
    :type min_content_length: int
    :param max_content_length: skip records with Content-Length large than this
    :type max_content_length: int
    :param func_filter: Python callable taking a :class:`WarcRecord` and returning a ``bool``
                        for further record filtering
    :type func_filter: Callable
    :param verify_digests: skip records which have no or an invalid block digest
    :type verify_digests: bool
    :param strict_mode: enforce strict spec compliance (setting this to ``False`` will enable quirks such
                        as ``LF`` instead of ``CRLF`` for headers)
    :type strict_mode: bool
    :param auto_decode: automatically decode record body if Content-Encoding or Transfer-Encoding headers are set
                        (accepted values: ``'none'`` [default], ``'content'``, ``'transfer'``, ``'all'``,
                        has no effect if ``parse_http`` is ``False``)
    :type auto_decode: str
    """

    def __init__(self, *args, **kwargs):
        pass

    def __cinit__(self, stream, uint16_t record_types=any_type, bint parse_http=True,
                  size_t min_content_length=strnpos, size_t max_content_length=strnpos,
                  func_filter=None, bint verify_digests=False, bint strict_mode=True,
                  str auto_decode='none'):
        self._set_stream(stream)
        self.record = None
        self.iter = None
        self.parse_http = parse_http
        self.verify_digests = verify_digests
        self.min_content_length = min_content_length
        self.max_content_length = max_content_length
        self.func_filter = func_filter
        self.record_type_filter = record_types
        self.strict_mode = strict_mode
        self.auto_decode = auto_decode

    def __iter__(self) -> t.Iterable[WarcRecord]:
        """
        __iter__(self)

        Iterate all :class:`WarcRecord` items in the current WARC stream.

        :rtype: t.Iterable[WarcRecord]
        """
        cdef _NextRecStatus status

        while True:
            status = self._read_next_record()
            if status == has_next:
                yield self.record
            elif status == eof:
                return

    def __next__(self) -> WarcRecord:
        """
        __next__(self)

        Implements an iterator that can be used with ``next()``.

        :rtype: WarcRecord
        """
        if self.iter is None:
            self.iter = self.__iter__()
        return next(self.iter)

    cdef _NextRecStatus _read_next_record(self) except _NextRecStatus.error:
        self.reader.detect_stream_type()

        if self.record is not None:
            self.reader.consume()
            self.reader.reset_limit()
            self.record._stale = not self.record._frozen

        self.record = WarcRecord.__new__(WarcRecord)
        self.record._stream_pos = self.reader.tell()

        cdef string version_line
        while True:
            version_line = self.reader.readline(self.strict_mode)

            if version_line.empty():
                # EOF
                return eof

            if version_line == b'\r\n' or version_line == b'\n':
                # Consume empty lines
                self.record._stream_pos = self.reader.tell()
                continue

            version_line = strip_str(move(version_line))
            if version_line.substr(0, 7) in [b'WARC/1.', b'WARC/0.']:
                # OK, continue with parsing headers
                self.record._headers.set_status_line(version_line)
                break
            else:
                # Not a WARC file or unsupported version
                return eof

        parse_header_block(self.reader, self.record._headers, False, self.strict_mode)

        cdef string hkey
        cdef size_t parse_count = 0
        cdef str_pair h
        cdef bint skip = False
        for h in self.record._headers._headers:
            hkey = str_to_lower(h[0])
            if hkey == b'warc-type':
                self.record._record_type = _str_record_type_to_enum(h[1])
                parse_count += 1
            elif hkey == b'content-length':
                self.record._content_length = strtol(h[1].c_str(), NULL, 10)
                parse_count += 1
            elif hkey == b'content-type' and h[1].find(b'application/http') == 0:
                self.record._is_http = True
                parse_count += 1

            if parse_count >= 3:
                break

        # Check if record is to be skipped
        skip |= (self.record._record_type & self.record_type_filter) == 0
        skip |= self.max_content_length != strnpos and self.record._content_length > self.max_content_length
        skip |= self.min_content_length != strnpos and self.record._content_length < self.min_content_length
        if not skip and self.func_filter is not None:
            # Execute expensive filters last and only if skip is not already true
            skip |= not self.func_filter(self.record)

        if skip:
            self.reader.consume(self.record._content_length)
            self.record = None
            return skip_next

        self.reader.set_limit(self.record._content_length)
        self.record._reader = self.reader

        if self.verify_digests and not self.record.verify_block_digest(False):
            self.reader.reset_limit()
            self.reader.consume(self.record._content_length)
            self.record = None
            return skip_next

        if self.parse_http and self.record._is_http:
            self.record.parse_http(self.strict_mode, self.auto_decode)

        return has_next

    cdef bint _set_stream(self, stream) except 0:
        """
        _set_stream(self, stream)

        Replace underlying input stream.

        This method is for internal use and should not be called by external users.
        """
        if not isinstance(stream, IOStream):
            for attr in ('read', 'tell', 'close'):
                if not hasattr(stream, attr):
                    raise AttributeError(f"Object of type '{type(stream).__name__}' has no attribute '{attr}'.")
            stream = PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, stream)

        cdef IOStream stream_ = <IOStream>stream
        self.reader = BufferedReader.__new__(BufferedReader, stream_)
        self.record = None
        return True


# noinspection PyProtectedMember
cpdef bint is_warc_10(WarcRecord record):
    """
    is_warc_10(record)

    Filter predicate for checking if record is a WARC/1.0 record.

    :param record: WARC record
    :type record: WarcRecord
    :rtype: bool
    """
    return record._headers._status_line == <char*>b'WARC/1.0'


# noinspection PyProtectedMember
cpdef bint is_warc_11(WarcRecord record):
    """
    is_warc_11(record)

    Filter predicate for checking if record is a WARC/1.1 record.

    :param record: WARC record
    :type record: WarcRecord
    :rtype: bool
    """
    return record._headers._status_line == <char*>b'WARC/1.1'


# noinspection PyProtectedMember
cpdef bint has_block_digest(WarcRecord record):
    """
    has_block_digest(record)

    Filter predicate for checking if record has a block digest.

    :param record: WARC record
    :type record: WarcRecord
    :rtype: bool
    """
    return not record._headers.find_header(<char*>b'WARC-Block-Digest', <char*>b'').empty()


# noinspection PyProtectedMember
cpdef bint has_payload_digest(WarcRecord record):
    """
    has_payload_digest(record)

    Filter predicate for checking if record has a payload digest.

    :param record: WARC record
    :type record: WarcRecord
    :rtype: bool
    """
    return not record._headers.find_header(<char*>b'WARC-Payload-Digest', <char*>b'').empty()


# noinspection PyProtectedMember
cpdef bint is_http(WarcRecord record):
    """
    is_http(record)

    Filter predicate for checking if record is an HTTP record.

    :param record: WARC record
    :type record: WarcRecord
    :rtype: bool
    """
    return record._is_http


# noinspection PyProtectedMember
cpdef bint is_concurrent(WarcRecord record):
    """
    is_concurrent(record)

    Filter predicate for checking if record is concurrent to another record.

    :param record: WARC record
    :type record: WarcRecord
    :rtype: bool
    """
    return not record._headers.find_header(<char*>b'WARC-Concurrent-To', <char*>b'').empty()



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

cimport cython
from cython.operator cimport preincrement as preinc
from libc.string cimport memchr, memcmp, memcpy
from libcpp.string cimport npos as strnpos, string

from resiliparse_inc.cstring cimport strerror
from resiliparse_inc.errno cimport errno
from resiliparse_inc.stdio cimport fclose, ferror, fflush, fopen, fread, fseek, ftell, fwrite, SEEK_SET

import brotli


class FastWARCError(Exception):
    """Generic FastWARC exception."""


class StreamError(FastWARCError):
    """FastWARC stream error."""


class ReaderStaleError(FastWARCError):
    """FastWARC reader stale error."""


@cython.auto_pickle(False)
cdef class IOStream:
    """IOStream base class."""

    def __enter__(self) -> IOStream:
        """
        :rtype: IOStream
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def read(self, size_t size):
        """
        read(self, size)

        Read ``size`` bytes from stream.

        :param size: bytes to read
        :type size: int
        :return: read bytes
        :rtype: bytearray
        """
        cdef bytearray buf = bytearray(size)
        size = self.read_(<char*>buf, size)
        return buf[:size]

    cdef size_t read_(self, char* out, size_t size) except -1:
        """Internal C interface for :meth:`read`."""
        pass

    cdef size_t write_(self, const char* data, size_t size) except -1:
        """Internal C interface for :meth:`write`."""
        pass

    def write(self, bytes data):
        """
        write(self, data)

        Write bytes to stream.

        :param data: data to write
        :type data: bytes
        :return: number of bytes written
        :rtype: int
        """
        size = self.write_(<const char*>data, len(data))
        return size

    cpdef void seek(self, size_t offset) except *:
        """
        seek(self, offset)
        
        Seek to specified offset.
        
        :param offset: seek offset
        :type offset: int
        """
        pass

    cpdef size_t tell(self) except -1:
        """
        tell(self)
        
        Return current stream offset.
        
        :return: stream offset
        :rtype: int
        """
        pass

    cpdef void flush(self) except *:
        """
        flush(self)
        
        Flush stream buffer.
        """
        pass

    cpdef void close(self) except *:
        """
        close(self)
        
        Close the stream.
        """
        pass


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class BytesIOStream(IOStream):
    """
    __init__(self, initial_data=None)

    IOStream that uses an in-memory buffer.

    :param initial_data: fill internal buffer with this initial data
    :type initial_data: bytes
    """
    def __init__(self, *args, **kwargs):
        pass

    def __cinit__(self, bytes initial_data=None):
        self.pos = 0
        if initial_data is not None:
            self.buffer = initial_data

    cpdef size_t tell(self) except -1:
        if self.pos == strnpos:
            raise ValueError('Trying I/O on closed file.')
        return self.pos

    cpdef void seek(self, size_t offset) except *:
        if self.pos == strnpos:
            raise ValueError('Trying I/O on closed file.')
        self.pos = min(self.buffer.size(), offset)

    cdef size_t read_(self, char* out, size_t size) except -1:
        if self.pos == strnpos:
            raise ValueError('Trying I/O on closed file.')
        if self.pos >= self.buffer.size():
            return 0
        if self.pos + size > self.buffer.size():
            size = self.buffer.size() - self.pos
        memcpy(out, self.buffer.data() + self.pos, size)
        self.seek(self.pos + size)
        return size

    cdef size_t write_(self, const char* data, size_t size) except -1:
        if self.pos == strnpos:
            raise ValueError('Trying I/O on closed file.')
        if self.pos + size > self.buffer.size():
            self.buffer.resize(self.pos + size)
        memcpy(self.buffer.data() + self.pos, data, size)
        self.pos += size
        return size

    cpdef void close(self) except *:
        self.buffer.clear()
        self.pos = strnpos

    cpdef string getvalue(self):
        """
        getvalue(self)
        
        Get buffer value.
        
        :return: buffer value
        :rtype: bytes
        """
        return self.buffer


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class FileStream(IOStream):
    """
    __init__(self, filename=None, mode='rb')

    Fast alternative to Python file objects for local files.

    :param filename: input filename
    :type filename: str
    :param mode: file open mode
    :type mode: str
    """

    def __init__(self, *args, **kwargs):
        pass

    def __cinit__(self, str filename, str mode='rb'):
        self.fp = NULL
        if filename:
            self.open_(filename.encode(), mode.encode())

    def __dealloc__(self):
        self.close()

    cdef void open_(self, char* path, char* mode=b'rb') except *:
        if self.fp != NULL:
            self.close()

        self.fp = fopen(path, mode)
        if self.fp == NULL:
            raise StreamError(strerror(errno).decode())

    cpdef void seek(self, size_t offset) except *:
        if self.fp == NULL:
            raise ValueError('Trying I/O on closed file.')
        fseek(self.fp, <unsigned long>offset, SEEK_SET)

    cpdef size_t tell(self) except -1:
        if self.fp == NULL:
            raise ValueError('Trying I/O on closed file.')
        return ftell(self.fp)

    cdef size_t read_(self, char* out, size_t size) except -1:
        if self.fp == NULL:
            raise ValueError('Trying I/O on closed file.')
        cdef size_t bytes_read
        with nogil:
            bytes_read = fread(out, sizeof(char), size, self.fp)
            if bytes_read < size and ferror(self.fp):
                with gil:
                    raise StreamError('Error reading file')
            return bytes_read

    cdef size_t write_(self, const char* data, size_t size) except -1:
        if self.fp == NULL:
            raise ValueError('Trying I/O on closed file.')
        cdef size_t w = fwrite(data, sizeof(char), size, self.fp)
        if ferror(self.fp):
            raise StreamError('Error writing file')
        return w

    cpdef void flush(self) except *:
        if self.fp == NULL:
            raise ValueError('Trying I/O on closed file.')
        fflush(self.fp)

    cpdef void close(self) except *:
        if self.fp != NULL:
            fclose(self.fp)
            self.fp = NULL


@cython.auto_pickle(False)
cdef class PythonIOStreamAdapter(IOStream):
    """
    __init__(self, py_stream)

    IOStream adapter for file-like Python objects.

    :param py_stream: input Python stream object
    """

    def __init__(self, *args, **kwargs):
        pass

    def __cinit__(self, py_stream):
        self.py_stream = py_stream

    cdef size_t read_(self, char* out, size_t size) except -1:
        cdef bytes data = self.py_stream.read(size)
        size = min(<size_t>len(data), size)
        memcpy(out, <char*>data, size * sizeof(char))
        return size

    cdef size_t write_(self, const char* data, size_t size) except -1:
        return self.py_stream.write(data[:size])

    cpdef size_t tell(self) except -1:
        return self.py_stream.tell()

    cpdef void seek(self, size_t offset) except *:
        self.py_stream.seek(offset)

    cpdef void flush(self) except *:
        self.py_stream.flush()

    cpdef void close(self) except *:
        if self.py_stream is None:
            return
        try:
            self.py_stream.close()
        except ValueError:
            # Ignore if already closed
            pass


cpdef IOStream wrap_stream(raw_stream):
    """
    wrap_stream(raw_stream)
    
    Wrap ``raw_stream`` into a :class:`PythonIOStreamAdapter` if it is a Python object or
    return ``raw_stream`` unmodified if it is a :class:`IOStream` already.
    
    :param raw_stream: stream to wrap
    :return: wrapped stream
    :rtype: IOStream
    """
    if isinstance(raw_stream, IOStream):
        return raw_stream
    elif isinstance(raw_stream, object) and hasattr(raw_stream, 'read'):
        return PythonIOStreamAdapter.__new__(PythonIOStreamAdapter, raw_stream)
    else:
        raise ValueError(f"Object of type '{type(raw_stream).__name__}' is not a valid stream.")


@cython.auto_pickle(False)
cdef class CompressingStream(IOStream):
    """Base class for compressed :class:`IOStream` types."""

    cpdef size_t begin_member(self):
        """
        begin_member(self)
        
        Begin compression member / frame (if not already started).
        
        :return: bytes written
        :rtype: int
        """
        return 0

    cpdef size_t end_member(self):
        """
        end_member(self)
        
        End compression member / frame (if one has been started).
        
        :return: bytes written
        :rtype: int
        """
        return 0


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class GZipStream(CompressingStream):
    """
    __init__(self, raw_stream, compression_level=9, zlib=False)

    GZip :class:`IOStream` implementation.

    :param raw_stream: raw data stream
    :param compression_level: GZip compression level (for compression only)
    :type compression_level: int
    :param zlib: use raw deflate / zlib format instead of gzip
    :type zlib: bool
    """

    def __init__(self, *args, **kwargs):
        pass

    def __cinit__(self, raw_stream, compression_level=Z_BEST_COMPRESSION, bint zlib=False):
        self.raw_stream = wrap_stream(raw_stream)
        self.member_started = False
        self.working_buf_filled = 0u
        self.stream_state = CompressingStreamState.UNINIT
        self.stream_pos = self.raw_stream.tell()
        self.compression_level = compression_level
        self.window_bits = MAX_WBITS if zlib else 16 + MAX_WBITS

    def __dealloc__(self):
        self.close()

    cpdef size_t tell(self) except -1:
        return self.stream_pos

    cdef void _init_z_stream(self, bint deflate) nogil:
        """
        Reset internal state and initialize ``z_stream``.
        
        :param deflate: ``True`` for compression context, ``False`` for decompression context.
        """

        if self.stream_state != CompressingStreamState.UNINIT:
            return

        self.zst.opaque = Z_NULL
        self.zst.zalloc = Z_NULL
        self.zst.zfree = Z_NULL
        self.zst.next_in = NULL
        self.zst.next_out = NULL
        self.zst.total_in = 0u
        self.zst.total_out = 0u
        self.zst.avail_in = 0u
        self.zst.avail_out = 0u
        self.working_buf.clear()

        if deflate:
            deflateInit2(&self.zst, self.compression_level, Z_DEFLATED, self.window_bits, 9, Z_DEFAULT_STRATEGY)
            self.stream_state = CompressingStreamState.COMPRESSING
        else:
            inflateInit2(&self.zst, self.window_bits)
            self.stream_state = CompressingStreamState.DECOMPRESSING

    cpdef void prepopulate(self, bint deflate, const string& initial_data):
        """
        prepopulate(self, initial_data)
        
        Fill internal working buffer with initial data.
        Use if some initial data of the stream have already been consumed (e.g., for stream content negotiation).
        Has to be called before the first :meth:`read()`.
        
        :param deflate: ``True`` if ``data`` is uncompressed, ``False`` if ``data`` is compressed GZip data. 
        :type deflate: int
        :param initial_data: data to pre-populate
        :type initial_data: bytes
        """
        self._init_z_stream(deflate)
        self.working_buf.append(initial_data)
        self.working_buf_filled += <unsigned int>initial_data.size()
        self.zst.next_in = <Bytef*>self.working_buf.data()
        self.zst.avail_in = self.working_buf_filled
        self.stream_pos = max(0u, self.raw_stream.tell() - self.working_buf_filled)

    cdef void _free_z_stream(self) nogil:
        """Release internal state and reset working buffer."""
        if self.stream_state == CompressingStreamState.UNINIT:
            return
        if self.stream_state == CompressingStreamState.COMPRESSING:
            deflateEnd(&self.zst)
        elif self.stream_state == CompressingStreamState.DECOMPRESSING:
            inflateEnd(&self.zst)
        self.working_buf.clear()
        self.working_buf_filled = 0u
        self.stream_state = CompressingStreamState.UNINIT

    cdef void _reinit_z_stream(self, bint deflate) nogil:
        """Re-initialize zstream, but retain working buffer. Use to restart stream with different window parameters."""

        cdef string working_buf_tmp = self.working_buf.substr(0, self.working_buf_filled)
        if self.stream_state != CompressingStreamState.UNINIT:
            self.zst.next_out = NULL
            self.zst.avail_out = 0u
            self._free_z_stream()
        with gil:
            self.prepopulate(deflate, working_buf_tmp)

    cdef bint _refill_working_buf(self, size_t size) except -1:
        if self.zst.avail_in > 0:
            return True

        if self.working_buf.size() < size:
            self.working_buf.resize(size)

        self.working_buf_filled = <unsigned int>self.raw_stream.read_(self.working_buf.data(), size)
        if self.working_buf_filled == 0:
            # EOF
            self._free_z_stream()
            return False

        self.zst.next_in = <Bytef*>self.working_buf.data()
        self.zst.avail_in = self.working_buf_filled
        return True

    cdef size_t read_(self, char* out, size_t size) except -1:
        if self.stream_state == CompressingStreamState.COMPRESSING:
            raise StreamError('Compression in progress.')

        # Init stream if this is the first read
        if self.stream_state == CompressingStreamState.UNINIT:
            self._init_z_stream(False)

            # If reader is in deflate / zlib format mode, check if first bytes are a zlib header
            if self.window_bits == MAX_WBITS:
                if not self._refill_working_buf(size) or self.working_buf.size() < 2:
                    return 0
                if not (self.working_buf[0] == b'\x78' and self.working_buf[1] in [b'\x01', b'\xef', b'\x9c', b'\xda']):
                    # Re-init as "raw" deflate stream if no zlib header found
                    self.window_bits = -MAX_WBITS
                    self._reinit_z_stream(False)

        if self.zst.avail_in == 0 or self.working_buf.empty():
            if not self._refill_working_buf(size):
                return 0

        self.zst.next_out = <Bytef*>out
        self.zst.avail_out = <unsigned int>size
        cdef int stream_read_status = Z_STREAM_END
        cdef size_t read_so_far = self.zst.total_in

        while True:
            stream_read_status = inflate(&self.zst, Z_NO_FLUSH)
            if self.zst.avail_out == 0 or (stream_read_status != Z_OK and stream_read_status != Z_BUF_ERROR):
                break
            if self.zst.avail_in == 0 and not self._refill_working_buf(size):
                break
        self.stream_pos += self.zst.total_in - read_so_far

        # Error
        if stream_read_status < 0 and stream_read_status != Z_BUF_ERROR:
            self._free_z_stream()
            raise StreamError('Not a valid GZip stream')

        # Member end
        if stream_read_status == Z_STREAM_END:
            inflateReset(&self.zst)

        return size - self.zst.avail_out

    cdef size_t write_(self, const char* data, size_t size) except -1:
        if self.stream_state == CompressingStreamState.DECOMPRESSING:
            raise StreamError('Decompression in progress')

        if not self.stream_state:
            self._init_z_stream(True)

        self.zst.next_in = <Bytef*>data
        self.zst.avail_in = <unsigned int>size

        self.begin_member()
        cdef size_t written = 0
        cdef size_t bound = max(8192u, deflateBound(&self.zst, <unsigned long>size))
        if self.working_buf.size() < bound:
            self.working_buf.resize(bound)
        self.zst.next_out = <Bytef*>self.working_buf.data()
        self.zst.avail_out = <unsigned int>self.working_buf.size()

        cdef int status = Z_OK
        cdef size_t written_so_far = self.zst.total_out

        while self.zst.avail_in > 0 and status == Z_OK:
            status = deflate(&self.zst, Z_NO_FLUSH)
            if self.zst.avail_in > 0 and self.zst.avail_out == 0:
                # Out buffer fully filled, but in buffer still holding data
                self.working_buf.resize(self.working_buf.size() + 4096u)
                self.zst.next_out = <Bytef*>self.working_buf.data() + self.working_buf.size() - 4096u
                self.zst.avail_out = 4096u

        written += self.zst.total_out - written_so_far
        self.stream_pos += written
        if written == 0:
            return 0

        return self.raw_stream.write_(self.working_buf.data(), written)

    cpdef size_t begin_member(self):
        self.member_started = True
        return 0

    cpdef size_t end_member(self):
        if not self.member_started:
            return 0

        self.zst.avail_in = 0u
        self.zst.next_in = NULL
        self.zst.next_out = <Bytef*>self.working_buf.data()
        self.zst.avail_out = <unsigned int>self.working_buf.size()

        cdef size_t written_so_far = self.zst.total_out
        cdef int status = deflate(&self.zst, Z_FINISH)
        while status == Z_OK or status == Z_BUF_ERROR:
            if self.zst.avail_out == 0:
                # Need larger output buffer (unlikely to ever happen at this point)
                self.working_buf.resize(self.working_buf.size() + 4096u)
                self.zst.next_out = <Bytef*>self.working_buf.data() + self.working_buf.size() - 4096u
                self.zst.avail_out = 4096u
            status = deflate(&self.zst, Z_FINISH)

        cdef size_t written = self.zst.total_out - written_so_far
        deflateReset(&self.zst)
        self.stream_pos += written
        self.member_started = False

        if written == 0:
            return 0
        return self.raw_stream.write_(self.working_buf.data(), written)

    cpdef void flush(self) except *:
        self.end_member()
        self.raw_stream.flush()

    cpdef void close(self) except *:
        self.end_member()
        self._free_z_stream()


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class LZ4Stream(CompressingStream):
    """
    __init__(self, raw_stream, compression_level=12, favor_dec_speed=True)

    LZ4 :class:`IOStream` implementation.

    :param raw_stream: raw data stream
    :param compression_level: LZ4 compression level (for compression only)
    :type compression_level: int
    :param favor_dec_speed: favour decompression speed over compression speed and size
    :type favor_dec_speed: bool
    """

    def __init__(self, *args, **kwargs):
        pass

    def __cinit__(self, raw_stream, compression_level=LZ4HC_CLEVEL_MAX, favor_dec_speed=True):
        self.raw_stream = wrap_stream(raw_stream)
        self.cctx = NULL
        self.dctx = NULL
        self.working_buf = string()
        self.working_buf_read = 0u
        self.working_buf_filled = 0u
        self.frame_started = False
        self.prefs.compressionLevel = compression_level
        self.prefs.favorDecSpeed = favor_dec_speed
        self.stream_pos = self.raw_stream.tell()

    def __dealloc__(self):
        self.close()

    cpdef size_t tell(self) except -1:
        return self.stream_pos

    cpdef void prepopulate(self, const string& initial_data):
        """
        prepopulate(self, initial_data)
        
        Fill internal working buffer with initial data.
        Use if some initial data of the stream have already been consumed (e.g., for stream content negotiation).
        Has to be called before the first :meth:`read()`.
        
        :param initial_data: data to pre-populate
        :type initial_data: bytes
        """
        self.working_buf.append(initial_data)
        self.working_buf_filled += <unsigned int>initial_data.size()
        self.stream_pos = max(0u, self.raw_stream.tell() - self.working_buf_filled)

    cdef size_t read_(self, char* out, size_t size) except -1:
        if self.cctx != NULL:
            raise StreamError('Compression in progress.')

        cdef size_t ret
        cdef size_t bytes_in = 0, bytes_out = 0
        cdef size_t out_buf_consumed = 0

        if self.dctx == NULL:
            LZ4F_createDecompressionContext(&self.dctx, LZ4F_VERSION)

        if self.working_buf.size() < size:
            self.working_buf.resize(size)

        while True:
            if self.working_buf_filled == 0 or self.working_buf_read == self.working_buf_filled:
                self.working_buf_filled = <unsigned int>self.raw_stream.read_(self.working_buf.data(), size)
                self.working_buf_read = 0

                if self.working_buf_filled == 0:
                    # EOF
                    self._free_ctx()
                    return out_buf_consumed

            bytes_in = self.working_buf_filled - self.working_buf_read
            bytes_out = size - out_buf_consumed
            ret = LZ4F_decompress(self.dctx, out + out_buf_consumed, &bytes_out,
                                  self.working_buf.data() + self.working_buf_read, &bytes_in, NULL)
            self.stream_pos += bytes_in
            self.working_buf_read += bytes_in
            out_buf_consumed += bytes_out

            if ret == 0 or out_buf_consumed == size or LZ4F_isError(ret):
                break

        if LZ4F_isError(ret):
            self._free_ctx()
            raise StreamError(f'Not a valid LZ4 stream: {LZ4F_getErrorName(ret).decode()}')

        return out_buf_consumed

    cpdef size_t begin_member(self):
        if self.frame_started:
            return 0

        cdef size_t ret = LZ4F_createCompressionContext(&self.cctx, LZ4F_VERSION)
        if LZ4F_isError(ret):
            raise StreamError(f'Failed to create compression context: {LZ4F_getErrorName(ret).decode()}')

        if self.working_buf.size() < LZ4F_HEADER_SIZE_MAX:
            self.working_buf.resize(LZ4F_HEADER_SIZE_MAX + 8192u)
        cdef size_t written = LZ4F_compressBegin(self.cctx, self.working_buf.data(),
                                                 self.working_buf.size(), &self.prefs)
        self.frame_started = True

        return self.raw_stream.write_(self.working_buf.data(), written)

    cpdef size_t end_member(self):
        if self.cctx == NULL or not self.frame_started:
            return 0

        cdef size_t written
        cdef size_t buf_needed = LZ4F_compressBound(0, &self.prefs)
        with nogil:
            if self.working_buf.size() < buf_needed:
                self.working_buf.resize(buf_needed)
            written = LZ4F_compressEnd(self.cctx, self.working_buf.data(), self.working_buf.size(), NULL)
            self.frame_started = False
        return self.raw_stream.write_(self.working_buf.data(), written)

    cdef size_t write_(self, const char* data, size_t size) except -1:
        if self.dctx != NULL:
            raise StreamError('Decompression in progress.')

        cdef size_t buf_needed = max(8192u, LZ4F_compressBound(size, &self.prefs))
        if self.working_buf.size() < buf_needed:
            self.working_buf.resize(buf_needed)

        cdef size_t header_bytes_written = self.begin_member()
        cdef size_t written = LZ4F_compressUpdate(self.cctx, self.working_buf.data(), self.working_buf.size(),
                                                  data, size, NULL)
        self.stream_pos += written
        return self.raw_stream.write_(self.working_buf.data(), written) + header_bytes_written

    cpdef void flush(self) except *:
        if self.cctx == NULL:
            return

        cdef size_t buf_needed = LZ4F_compressBound(0, &self.prefs)
        if self.working_buf.size() < buf_needed:
            self.working_buf.resize(buf_needed)

        cdef size_t written = LZ4F_flush(self.cctx, self.working_buf.data(), self.working_buf.size(), NULL)
        self.raw_stream.write_(self.working_buf.data(), written)
        self.stream_pos += written
        self.raw_stream.flush()

    cpdef void close(self) except *:
        if self.cctx != NULL:
            self.end_member()

        self._free_ctx()

    cdef void _free_ctx(self) nogil:
        if self.cctx != NULL:
            LZ4F_freeCompressionContext(self.cctx)
            self.cctx = NULL

        if self.dctx != NULL:
            LZ4F_freeDecompressionContext(self.dctx)
            self.dctx = NULL

        if not self.working_buf.empty():
            self.working_buf.clear()
        self.working_buf_filled = 0u



# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class BrotliStream(CompressingStream):
    """
    __init__(self, raw_stream, quality=11, lgwin=22, lgblock=0)

    Brotli :class:`IOStream` implementation.

    Implementation relies on Google's ``brotli`` Python package, will be ported to native
    C version in a later version.

    :param raw_stream: raw data stream
    :param quality: compression quality (higher quality means better compression, but less speed)
    :type quality: int
    :param lgwin: Base 2 logarithm of the sliding window size in the range 16 to 24
    :type lgwin: int
    :param lgblock: Base 2 logarithm of the maximum input block size in the range 16 to 24
                    (will be set based on quality of value is 0)
    :type lgblock: int
    """
    def __init__(self, *args, **kwargs):
        pass

    def __cinit__(self, raw_stream, size_t quality=11, size_t lgwin=22, size_t lgblock=0):
        self.raw_stream = raw_stream
        self.quality = quality
        self.lgwin = lgwin
        self.lgblock = lgblock
        self.stream_state = CompressingStreamState.UNINIT
        self.compressor = None

    cpdef size_t begin_member(self):
        pass

    cpdef size_t end_member(self):
        pass

    cdef size_t read_(self, char* out, size_t size) except -1:
        if self.stream_state == CompressingStreamState.COMPRESSING:
            raise StreamError('Compression in progress.')
        elif self.stream_state == 0:
            self.compressor = brotli.Decompressor()
            self.stream_state = CompressingStreamState.DECOMPRESSING

        cdef string tmp_buffer
        tmp_buffer.resize(size)
        cdef size_t bytes_read = self.raw_stream.read_(tmp_buffer.data(), size)
        tmp_buffer.resize(bytes_read)

        try:
            self.working_buf.append(<string>self.compressor.process(tmp_buffer))
        except brotli.error:
            raise StreamError(f'Not a valid Brotli stream')

        bytes_read = min(self.working_buf.size(), size)
        memcpy(out, self.working_buf.data(), bytes_read)
        self.working_buf = self.working_buf.substr(bytes_read)
        return bytes_read


    cdef size_t write_(self, const char* data, size_t size) except -1:
        if self.stream_state == CompressingStreamState.DECOMPRESSING:
            raise StreamError('Decompression in progress.')
        elif self.stream_state == 0:
            self.compressor = brotli.Compressor()
            self.stream_state = CompressingStreamState.COMPRESSING
        cdef bytes compressed = self.compressor.process(<bytes>data[:size])
        return self.raw_stream.write_(<char*>compressed, len(compressed))

    cpdef size_t tell(self) except -1:
        return self.raw_stream.tell()

    cpdef void seek(self, size_t offset) except *:
        self.raw_stream.seek(offset)

    cpdef void flush(self)  except *:
        if self.stream_state == CompressingStreamState.COMPRESSING:
            compressed = self.compressor.flush()
            self.raw_stream.write_(<char*>compressed, len(compressed))
            self.raw_stream.flush()

    cpdef void close(self) except *:
        if self.stream_state == CompressingStreamState.COMPRESSING:
            compressed = self.compressor.finish()
            self.raw_stream.write_(<char*>compressed, len(compressed))
        self.compressor = None
        self.stream_state = CompressingStreamState.UNINIT
        self.working_buf.clear()


# noinspection PyAttributeOutsideInit
@cython.auto_pickle(False)
cdef class BufferedReader:
    """
    __init__(self, stream, buf_size=16384, negotiate_stream=True)

    Buffered reader operating on an :class:`IOStream` instance.

    :param stream: stream to operate on
    :type stream: IOStream
    :param buf_size: internal buffer size
    :type buf_size: int
    :param negotiate_stream: whether to auto-negotiate stream type
    :type negotiate_stream: bool
    """

    def __init__(self, *args, **kwargs):
        pass

    def __cinit__(self, IOStream stream, size_t buf_size=65536, bint negotiate_stream=True):
        self.stream = stream
        self.buf = string()
        self.buf.resize(max(4096u, buf_size))
        self.buf_view = string_view()
        self.limited_buf_view = string_view()
        self.limit = strnpos
        self.limit_consumed = 0
        self.stream_is_compressed = isinstance(stream, CompressingStream)
        self.stream_started = False
        self.negotiate_stream = negotiate_stream and not self.stream_is_compressed

    # noinspection PyAttributeOutsideInit
    cdef bint detect_stream_type(self) except 0:
        """
        Try to auto-detect stream type (GZip, LZ4, or uncompressed).
        """
        if not self.negotiate_stream or self.stream_started:
            return True

        self.negotiate_stream = False
        self._fill_buf()

        if self.buf_view.size() > 2 and self.buf_view[0] == <char>0x1f and self.buf_view[1] == <char>0x8b:
            self.stream = GZipStream.__new__(GZipStream, self.stream)
            (<GZipStream> self.stream).prepopulate(False, <string>self.buf_view)
        elif self.buf_view.size() > 4 and memcmp(self.buf_view.data(), <const char*>b'\x04\x22\x4d\x18', 4) == 0:
            self.stream = LZ4Stream.__new__(LZ4Stream, self.stream)
            (<LZ4Stream> self.stream).prepopulate(<string>self.buf_view)
        elif self.buf_view.size() > 5 and memcmp(self.buf_view.data(), <const char*>b'WARC/', 5) == 0:
            # Stream is uncompressed: bail out, don't mess with buffers
            self.stream_is_compressed = False
            return True
        else:
            self.stream_is_compressed = False
            raise StreamError('Not a valid WARC stream')

        self.stream_is_compressed = isinstance(self.stream, CompressingStream)
        self.buf_view.remove_prefix(self.buf_view.size())
        self.stream_started = False
        return True

    cdef bint _fill_buf(self) except -1:
        """
        Refill internal buffer.
        
        :return: ``True`` if refill was successful, ``False`` otherwise (EOF)
        """
        self.stream_started = True
        if self.buf_view.size() > 0:
            return True if self.limit == strnpos else self.limit > self.limit_consumed

        cdef size_t bytes_read = self.stream.read_(self.buf.data(), self.buf.size())
        self.buf_view = string_view(self.buf.data(), bytes_read)

        if self.buf_view.size() == 0:
            return False
        elif self.limit != strnpos:
            return self.limit > self.limit_consumed

        return True

    cdef string_view* _get_buf(self) nogil:
        """
        Get buffer contents. Does take a set limit into account.
        
        Returns a pointer, since Cython does not support returning lvalue references.
        
        :return: available buffer contents
        """
        if self.limit == strnpos:
            return &self.buf_view

        cdef size_t remaining = self.limit - self.limit_consumed
        self.limited_buf_view = string_view(self.buf_view.data(), self.buf_view.size())
        if self.limited_buf_view.size() > remaining:
            self.limited_buf_view.remove_suffix(self.limited_buf_view.size() - remaining)
        return &self.limited_buf_view

    cdef size_t _consume_buf(self, size_t size) nogil:
        """
        Consume up to ``size`` bytes from internal buffer. Takes a set limit into account.
        
        :param size: number of bytes to read
        :return: bytes consumed
        """
        cdef size_t consumed
        if self.limit == strnpos and size >= self.buf_view.size():
            consumed = self.buf_view.size()
            self.buf_view.remove_prefix(consumed)
            return consumed

        if self.limit != strnpos:
            if size > self.limit - self.limit_consumed:
                size = self.limit - self.limit_consumed
            self.limit_consumed += size

        if size > self.buf_view.size():
            size = self.buf_view.size()
        self.buf_view.remove_prefix(size)
        return size

    cdef inline void set_limit(self, size_t offset) nogil:
        """
        Set a stream limit in bytes. Any read beyond this limit will act as if the stream reached EOF.
        A set limit can be reset by calling :meth:`reset_limit()`.
        
        :param offset: limit in bytes
        """
        self.limit = offset
        self.limit_consumed = 0

    cdef inline void reset_limit(self) nogil:
        """Reset any previously set stream limit."""
        self.limit = strnpos

    cpdef string read(self, size_t size=strnpos) except *:
        """
        read(self, size=-1)
        
        Read up to ``size`` bytes from the input stream.
        
        :param size: number of bytes to read (default means read remaining stream)
        :type size: int
        :return: consumed buffer contents as bytes (or empty string if EOF)
        :rtype: bytes
        """
        cdef string data_read
        cdef size_t remaining = size
        cdef string_view buf_sub

        while (size == strnpos or data_read.size() < size) and self._fill_buf():
            buf = self._get_buf()
            remaining = size - data_read.size()
            buf_sub = self._get_buf().substr(0, remaining)
            data_read.append(buf_sub.data(), buf_sub.size())
            self._consume_buf(buf_sub.size())
        return data_read

    cpdef string readline(self, bint crlf=True, size_t max_line_len=8192) except *:
        """
        readline(self, crlf=True, max_line_len=8192)
        
        Read a single line from the input stream.
        
        :param crlf: whether lines are separated by CRLF or LF
        :type crlf: bool
        :param max_line_len: maximum line length (longer lines will still be consumed, but the return
                             value will not be larger than this)
        :type max_line_len: int
        :return: line contents (or empty string if EOF)
        :rtype: bytes
        """

        cdef string_view* buf
        cdef size_t capacity_remaining = max_line_len
        cdef bint last_was_cr = False

        cdef string line
        line.reserve(192)

        cdef size_t lf_pos = 0
        cdef char* lf_ptr = NULL

        while lf_ptr == NULL and self._fill_buf():
            buf = self._get_buf()

            if crlf:
                if last_was_cr and buf.front() == b'\n':
                    line.push_back(b'\n')
                    self._consume_buf(1)
                    return line
                last_was_cr = False

                lf_ptr = <char*>memchr(buf.data(), <int>b'\r', buf.size())
                while lf_ptr != NULL:
                    lf_pos = lf_ptr - buf.data() + 1
                    if lf_pos == buf.size():
                        last_was_cr = True
                        lf_ptr = NULL
                        break
                    preinc(lf_ptr)
                    if lf_ptr[0] == b'\n':
                        break
                    lf_ptr = <char*>memchr(lf_ptr, <int>b'\r', buf.size() - lf_pos)
            else:
                lf_ptr = <char*>memchr(buf.data(), <int>b'\n', buf.size())
                if lf_ptr != NULL:
                    lf_pos = lf_ptr - buf.data()

            if lf_ptr == NULL:
                lf_pos = buf.size() - 1

            preinc(lf_pos)
            line.append(buf.data(), min(lf_pos, capacity_remaining))
            capacity_remaining = max_line_len - line.size()
            self._consume_buf(lf_pos)

        return line

    cpdef size_t tell(self) except -1:
        """
        tell(self)
        
        Offset on the input stream.
        
        :return: offset
        :rtype: int
        """
        if not self.stream_is_compressed and not self.stream_started:
            return 0

        if self.limit != strnpos:
            return self.limit_consumed

        if self.stream_is_compressed:
            # Use position as returned by the decompressor.
            # The BufferedReader position inside a decompressed stream is meaningless.
            return self.stream.tell()

        return self.stream.tell() - self.buf_view.size()

    cpdef size_t consume(self, size_t size=strnpos) except -1:
        """
        consume(self, size=-1)
        
        Consume up to ``size`` bytes from the input stream without allocating a buffer for it.
        
        :param size: number of bytes to read (default means read remaining stream)
        :type size: int
        :return: number of bytes consumed
        :rtype: int
        """
        cdef string_view* buf
        cdef size_t consumed = 0

        while size > consumed and self._fill_buf():
            buf = self._get_buf()
            if buf.empty():
                break

            if size != strnpos:
                consumed += self._consume_buf(size - consumed)
            else:
                consumed += self._consume_buf(buf.size())

        return consumed

    cpdef void close(self) except *:
        """
        close(self)
        
        Close stream.
        """
        if self.stream is not None:
            self.stream.close()


def _buf_reader_py_test_detect_stream_type(BufferedReader buf):
    """
    _buf_reader_py_test_detect_stream_type(buf):

    Test interface for :meth:`BufferedReader.detect_stream_type`
    """
    buf.detect_stream_type()


def _buf_reader_py_test_set_limit(BufferedReader buf, size_t limit):
    """
    _buf_reader_py_test_detect_set_limit(buf, limit):

    Test interface for :meth:`BufferedReader.set_limit`
    """
    buf.set_limit(limit)


def _buf_reader_py_test_reset_limit(BufferedReader buf):
    """
    _buf_reader_py_test_reset_limit(buf, limit):

    Test interface for :meth:`BufferedReader.reset_limit`
    """
    buf.reset_limit()



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain
import os
import platform
import sys

import distutils.ccompiler
from distutils.dir_util import copy_tree
from setuptools import find_packages, setup, Extension


# 1. BOILERPLATE -------------------------------------------------------

VERSION = '0.14.1'
ROOT_DIRECTORY = os.path.abspath(os.path.dirname(__file__))
CXX = distutils.ccompiler.get_default_compiler()

TRACE = bool(int(os.getenv('TRACE', 0)))
DEBUG = bool(int(os.getenv('DEBUG', 0))) or TRACE
ASAN = bool(int(os.getenv('ASAN', 0)))

cpp_args = {}
try:
    from Cython.Build import cythonize
    import Cython.Compiler.Options

    cpp_ext = 'pyx'
    cython_args = dict(
        annotate=DEBUG,
        gdb_debug=DEBUG,
        compiler_directives=dict(
            language_level='3',
            linetrace=TRACE,
            initializedcheck=DEBUG,
            boundscheck=DEBUG,
            cdivision=True
        )
    )
    USE_CYTHON = True
except ModuleNotFoundError as e:
    cpp_ext = 'cpp'
    cython_args = {}
    USE_CYTHON = False

if TRACE:
    cpp_args.update(dict(define_macros=[('CYTHON_TRACE_NOGIL', '1')]))

if CXX == 'unix':
    cpp_args.update(dict(
        extra_compile_args=['-std=c++17',
                            f'-O{0 if DEBUG else 3}',
                            '-Wall',
                            '-Wno-deprecated-declarations',
                            '-Wno-unreachable-code',
                            '-Wno-unused-function'],
        extra_link_args=['-std=c++17']
    ))
    if DEBUG:
        cpp_args['extra_compile_args'].append('-Werror')
    if ASAN:
        cpp_args['extra_compile_args'].append('-fsanitize=address')
        cpp_args['extra_link_args'].append('-fsanitize=address')
    if os.getenv('CIBUILDWHEEL'):
        # Programmatically set correct include and library paths for macOS CI cross-compilation
        arch = os.getenv('_PYTHON_HOST_PLATFORM', '').split('-')[-1] or platform.processor()
        arch = arch.replace('x86_64', 'x64')
        if arch in ['x64', 'arm64']:
            cpp_args['extra_compile_args'].append(f'-I/usr/local/share/vcpkg/installed/{arch}-osx/include')
            cpp_args['extra_link_args'].append(f'-L/usr/local/share/vcpkg/installed/{arch}-osx/lib')

elif CXX == 'msvc':
    cpp_args.update(dict(
        extra_compile_args=['/std:c++latest', '/W3'],
        extra_link_args=[]
    ))
    if DEBUG:
        cpp_args['extra_compile_args'].append('/WX')
        cpp_args['extra_link_args'].append('/WX')

if os.path.isdir(os.path.join(ROOT_DIRECTORY, '..', 'resiliparse_inc')):
    copy_tree(os.path.join(ROOT_DIRECTORY, '..', 'resiliparse_inc'),
              os.path.join(ROOT_DIRECTORY, 'resiliparse_inc'), update=1)
    copy_tree(os.path.join(ROOT_DIRECTORY, '..', 'resiliparse_common'),
              os.path.join(ROOT_DIRECTORY, 'resiliparse_common'), update=1)

data_ext = ['*.pxd', '*.h']
inc_package = []
if 'sdist' in sys.argv:
    # Include resiliparse_inc module and Cython src files only in source distribution
    data_ext.extend(['*.pyx', '*.pxi'])
    inc_package.append('resiliparse_inc')
    inc_package.append('resiliparse_common')


# 2. FASTWARC SETUP -------------------------------------------------------

fastwarc_extensions = [
    Extension('fastwarc.warc', sources=[f'fastwarc/warc.{cpp_ext}'], **cpp_args),
    Extension('fastwarc.stream_io', sources=[f'fastwarc/stream_io.{cpp_ext}'],
              libraries=['zlib' if CXX == 'msvc' else 'z', 'lz4'], **cpp_args),
    Extension('fastwarc.tools', sources=[f'fastwarc/tools.{cpp_ext}'], **cpp_args)
]
if USE_CYTHON:
    fastwarc_extensions = cythonize(fastwarc_extensions, **cython_args)

extras_require = {}
extras_require.update({
    'all': list(chain(*extras_require.values())),   # All except "test"
})

tests_require = [
    'pytest',
    'pytest-cov',
    'lz4'
]
extras_require['test'] = tests_require
setup(
    name='FastWARC',
    version=VERSION,
    description='A high-performance WARC parsing library for Python written in C++/Cython.',
    long_description=open(os.path.join(ROOT_DIRECTORY, 'README.md')).read(),
    long_description_content_type='text/markdown',
    author='Janek Bevendorff',
    url='https://github.com/chatnoir-eu/chatnoir-resiliparse',
    license='Apache License 2.0',
    packages=[*find_packages(), *inc_package],
    package_data={'': data_ext},
    ext_modules=fastwarc_extensions,
    install_requires=[
        'brotli',
        'click',
        'tqdm'
    ],
    setup_requires=['setuptools>=18.0'],
    tests_require=tests_require,
    extras_require=extras_require,
    entry_points={
        'console_scripts': ['fastwarc=fastwarc.cli:main']
    }
)



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport uint16_t
from libcpp.string cimport string
from libcpp.vector cimport vector

from fastwarc.stream_io cimport IOStream, BufferedReader


cpdef enum WarcRecordType:
    warcinfo = 2,
    response = 4,
    resource = 8,
    request = 16,
    metadata = 32,
    revisit = 64,
    conversion = 128,
    continuation = 256,
    unknown = 512,
    any_type = 65535,
    no_type = 0


ctypedef (string, string) str_pair


cdef class WarcHeaderMap:
    cdef string _status_line
    cdef vector[str_pair] _headers
    cdef str _enc
    cdef dict _dict_cache
    cdef bint _dict_cache_stale

    cpdef size_t write(self, IOStream stream) except -1
    cpdef void clear(self)
    cdef inline void set_status_line(self, const string& status_line)
    cdef string find_header(self, const string& header_key, const string& default)
    cdef void set_header(self, const string& header_key, const string& header_value)
    cdef inline void append_header(self, const string& header_key, const string& header_value)
    cdef void add_continuation(self, const string& header_continuation_value)


cdef enum _NextRecStatus:
    has_next,
    skip_next,
    eof,
    error


cdef class WarcRecord:
    cdef WarcRecordType _record_type
    cdef WarcHeaderMap _headers
    cdef bint _is_http
    cdef bint _http_parsed
    cdef string _http_charset
    cdef WarcHeaderMap _http_headers
    cdef size_t _content_length
    cdef BufferedReader _reader
    cdef size_t _stream_pos
    cdef bint _stale
    cdef bint _frozen

    cpdef void init_headers(self, size_t content_length=*, WarcRecordType record_type=*, bytes record_urn=*)
    cpdef void set_bytes_content(self, bytes b)
    cpdef bint parse_http(self, bint strict_mode=*, str auto_decode=*) except 0
    cdef bint _init_content_decoder(self, str decode_mode) except -1
    cpdef size_t write(self, stream, bint checksum_data=*, bytes payload_digest=*, size_t chunk_size=*) except -1
    cpdef bint freeze(self) except 0
    cpdef bint verify_block_digest(self, bint consume=*) except -1
    cpdef bint verify_payload_digest(self, bint consume=*) except -1

    cdef bint _verify_digest(self, const string& base32_digest, bint consume) except -1
    cdef size_t _write_impl(self, in_reader, out_stream, bint write_payload_headers, size_t chunk_size) except -1
    cdef inline bint _assert_not_stale(self) except 0


cdef class ArchiveIterator:
    cdef BufferedReader reader
    cdef WarcRecord record
    cdef iter
    cdef bint parse_http
    cdef bint verify_digests
    cdef size_t min_content_length
    cdef size_t max_content_length
    cdef object func_filter
    cdef uint16_t record_type_filter
    cdef bint strict_mode
    cdef str auto_decode

    cdef _NextRecStatus _read_next_record(self) except _NextRecStatus.error
    cdef bint _set_stream(self, stream) except 0


cpdef bint is_warc_10(WarcRecord record)
cpdef bint is_warc_11(WarcRecord record)
cpdef bint has_block_digest(WarcRecord record)
cpdef bint has_payload_digest(WarcRecord record)
cpdef bint is_http(WarcRecord record)
cpdef bint is_concurrent(WarcRecord record)
